#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <assert.h>
#include <stdio.h>
#include <Xm/ScrollBar.h>
#include "line3d.bm"
#include "FTextP.h"

static void activate( Widget, XEvent *, String *, Cardinal *);
static void motion( Widget, XEvent *, String *, Cardinal *);

static XtActionsRec actionsList[] =
{
    {"activate", activate},
    {"motion", motion},
};

static char defaultTranslations[] = "\
<Btn1Down>,<Btn1Up>: activate() \n\
<Motion>: motion() \n\
<Enter>: motion() \n\
<Leave>: motion() \n\
";

static void _resolve_inheritance( WidgetClass);
static void initialize( Widget, Widget, ArgList, Cardinal *);
static void destroy( Widget);
static void resize( Widget);
static void expose( Widget, XEvent *, Region);
static XtGeometryResult geometry_manager( Widget, XtWidgetGeometry *, XtWidgetGeometry *);
static XtGeometryResult query_geometry( Widget, XtWidgetGeometry *, XtWidgetGeometry *);
static void add_word( Widget, const String, Cardinal, Pixel, Pixel, TextStyle, XtPointer);
static void add_hspace( Widget, int, Bool, Bool, Pixel, Pixel, TextStyle, XtPointer);
static void add_vspace( Widget, int);
static void add_inline( Widget, Widget, TextStyle, int, int, int, int, int, XtPointer);
static void add_parshape( Widget, int, int, int, int, double);
static void add_hrule( Widget);
static void add_eod( Widget);
static void pass_click( Widget, Widget, int, int);
static void reformat_scrollbar( Widget);
static int get_em_of_textstyle( Widget, TextStyle);
static int get_space_of_textstyle( Widget, TextStyle);
static void get_lineheight_of_textstyle( Widget, TextStyle, int *, int *);
static void get_word_extent( Widget, const String, Cardinal, TextStyle, int *, int *, int *, int *);

#define PREF_WD 150
#define PREF_HT 150
#define new(p) ((p )=(XtPointer )XtMalloc (sizeof (*(p ))))
#define renewarray(p, n) ((p )=(XtPointer )XtRealloc ((String )p ,(n )*sizeof (*(p ))))

static void draw_word( Widget, FtChunk);
static void draw_hspace( Widget, FtChunk);
static void draw_inline( Widget, FtChunk);
static void draw_hrule( Widget, FtChunk);
static void draw_chunks( Widget, FtChunk, FtChunk);

#define set_max(a, b) if ((a )>=(b ));else (a )=(b )
#define set_min(a, b) if ((a )<=(b ));else (a )=(b )
#define min(a, b) ((a )<(b )?(a ):(b ))
#define max(a, b) ((a )>(b )?(a ):(b ))
#define round(f) ((int )((f )+0.5 ))

static void init_par_builder( Widget);
static void append( Widget, FtChunk);
static String strndup( const String, Cardinal);
static int find_line_below_y( FtLine *, int, int);
static FtChunk local_find_chunk_at_xy( FtLine *, int, int, int);
static FtChunk find_chunk_at_xy( Widget, int, int);
static void resort_lines_vertically( FtLine *, int);

#define INCR 50

static void append_line( Widget, int, int, int, int, int, FtChunk, FtChunk);
static FtChunk find_parshape_before( FtChunk);

#define is_breakable(h) ((h ->tp ==FtHSpace &&h ->u.hspace.breakable )||h ->tp ==FtVSpace ||h ->tp ==FtParShape ||h ->tp ==FtHRule )


static void render_one_line( Widget, FtChunk, FtChunk, int, int, int, int, int, int, int, int, Bool);
static FtChunk find_break( FtChunk, FtChunk, int *, int *, int *, int *);
static int left_float_width( Widget, int, int);
static int right_float_width( Widget, int, int);
static int compute_width( Widget, FtChunk, int, int);
static FtChunk render_paragraphs( Widget, FtChunk);
static void format_vspace( Widget, FtChunk);
static void format_inline( Widget, FtChunk);
static int find_width_before( Widget, FtChunk);
static void format_parshape( Widget, FtChunk);
static void format_hrule( Widget, FtChunk);
static void unmap_children( Widget);
static void scroll_cb( Widget, XtPointer, XtPointer);
static Boolean do_layout( XtPointer);
static void install_do_layout( Widget);
static void load_font( Widget, int, int, int);
static void interpret_style( Widget, TextStyle, int *, int *, int *, int *);
static void font_extents( Widget, TextStyle, const String, int, int *, int *, int *, int *);
static void font_space( Widget, TextStyle, int *, int *, int *, int *);
static Font font_id( Widget, TextStyle);
static void dump_chunk( FtChunk);
static void dump( Widget);


static void draw_word(Widget self, FtChunk p)
{
    int fam, sty, siz, raise, ulpos, ulthick, w;

    assert(p->tp == FtWord);
    XSetFont(XtDisplay(self), ((XfwfFormattedTextWidget) self)->xfwfFormattedText.gc, p->u.word.fid);
    XSetForeground(XtDisplay(self), ((XfwfFormattedTextWidget) self)->xfwfFormattedText.gc, p->u.word.fg);
    if (p->u.word.bg != TRANSPARENT)
	XSetBackground(XtDisplay(self), ((XfwfFormattedTextWidget) self)->xfwfFormattedText.gc, p->u.word.bg);

    if (p->u.word.bg == TRANSPARENT)
	XDrawString(XtDisplay(self), XtWindow(self), ((XfwfFormattedTextWidget) self)->xfwfFormattedText.gc,
		    p->x, p->y - ((XfwfFormattedTextWidget) self)->xfwfFormattedText.voffset, p->u.word.text, p->u.word.len);
    else
	XDrawImageString(XtDisplay(self), XtWindow(self), ((XfwfFormattedTextWidget) self)->xfwfFormattedText.gc,
			 p->x, p->y - ((XfwfFormattedTextWidget) self)->xfwfFormattedText.voffset, p->u.word.text, p->u.word.len);

    if (p->u.word.style & (FtUNDERSCORE | FtUNDERSCORE2)) {
	interpret_style(self, p->u.word.style, &fam, &sty, &siz, &raise);
	ulpos = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fonts[fam][sty][siz].ulpos;
	ulthick = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fonts[fam][sty][siz].ulthick;
	w = XTextWidth(((XfwfFormattedTextWidget) self)->xfwfFormattedText.fonts[fam][sty][siz].fs, p->u.word.text, p->u.word.len);
	if (w != 0 && ulthick != 0 && (p->u.word.style & FtUNDERSCORE))
	    XFillRectangle(XtDisplay(self), XtWindow(self), ((XfwfFormattedTextWidget) self)->xfwfFormattedText.gc,
			   p->x, p->y - ((XfwfFormattedTextWidget) self)->xfwfFormattedText.voffset + ulpos, w, ulthick);
	if (w != 0 && ulthick != 0 && (p->u.word.style & FtUNDERSCORE2))
	    XFillRectangle(XtDisplay(self), XtWindow(self), ((XfwfFormattedTextWidget) self)->xfwfFormattedText.gc,
			   p->x, p->y - ((XfwfFormattedTextWidget) self)->xfwfFormattedText.voffset + ulpos + 2, w, ulthick);
    }
}

static void draw_hspace(Widget self, FtChunk p)
{
    int fam, sty, siz, raise, ulpos, ulthick;

    /* assert(p->tp == FtHSpace); */
    if (p->w > 0 && p->u.hspace.bg != TRANSPARENT) {
	XSetForeground(XtDisplay(self), ((XfwfFormattedTextWidget) self)->xfwfFormattedText.spacegc, p->u.hspace.bg);
	XFillRectangle(XtDisplay(self), XtWindow(self), ((XfwfFormattedTextWidget) self)->xfwfFormattedText.spacegc, p->x,
		       p->y - ((XfwfFormattedTextWidget) self)->xfwfFormattedText.voffset - p->h, p->w, p->h + p->d);
    }
    if (p->w > 0 && (p->u.hspace.style & (FtUNDERSCORE | FtUNDERSCORE2))) {
	interpret_style(self, p->u.hspace.style, &fam, &sty, &siz, &raise);
	ulpos = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fonts[fam][sty][siz].ulpos;
	ulthick = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fonts[fam][sty][siz].ulthick;
	if (ulthick != 0 && (p->u.hspace.style & FtUNDERSCORE))
	    XFillRectangle(XtDisplay(self), XtWindow(self), ((XfwfFormattedTextWidget) self)->xfwfFormattedText.gc,
			   p->x, p->y - ((XfwfFormattedTextWidget) self)->xfwfFormattedText.voffset + ulpos, p->w, ulthick);
	if (ulthick != 0 && (p->u.hspace.style & FtUNDERSCORE2))
	    XFillRectangle(XtDisplay(self), XtWindow(self), ((XfwfFormattedTextWidget) self)->xfwfFormattedText.gc,
			   p->x, p->y - ((XfwfFormattedTextWidget) self)->xfwfFormattedText.voffset + ulpos + 2, p->w, ulthick);
    }
}

static void draw_inline(Widget self, FtChunk p)
{
    Widget w = p->u.inlined.widget;
    if (((XfwfFormattedTextWidget) w)->core.x != p->x + p->u.inlined.hmargin
	|| ((XfwfFormattedTextWidget) w)->core.y != p->y - p->h + p->u.inlined.vmargin - ((XfwfFormattedTextWidget) self)->xfwfFormattedText.voffset) {
	XtMoveWidget(w, p->x + p->u.inlined.hmargin,
		     p->y - p->h + p->u.inlined.vmargin - ((XfwfFormattedTextWidget) self)->xfwfFormattedText.voffset);
    }
    if (!p->u.inlined.mapped) {
	XtMapWidget(w);
	p->u.inlined.mapped = TRUE;
    }
}

static void draw_hrule(Widget self, FtChunk p)
{
    XSetTSOrigin(XtDisplay(self), ((XfwfFormattedTextWidget) self)->xfwfFormattedText.linegc, p->x, p->y - ((XfwfFormattedTextWidget) self)->xfwfFormattedText.voffset);
    XFillRectangle(XtDisplay(self), XtWindow(self), ((XfwfFormattedTextWidget) self)->xfwfFormattedText.linegc, p->x,
		   p->y - ((XfwfFormattedTextWidget) self)->xfwfFormattedText.voffset, p->w, p->d);
}

static void draw_chunks(Widget self, FtChunk first, FtChunk stop)
{
    for (; first != stop; first = first->next)
	switch (first->tp) {
	case FtWord:
	    draw_word(self, first);
	    break;
	case FtHSpace:
	    draw_hspace(self, first);
	    break;
	case FtVSpace:
	    break;
	case FtInline:
	    draw_inline(self, first);
	    break;
	case FtParShape:
	    break;
	case FtHRule:
	    draw_hrule(self, first);
	    break;
	default:
	    assert(!"Cannot happen");
	}
}

static void init_par_builder(Widget self)
{
    Widget vscroll = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.vscrollbar;
    ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fullwidth = ((XfwfFormattedTextWidget) self)->core.width - ((XfwfFormattedTextWidget) vscroll)->core.width;
    ((XfwfFormattedTextWidget) self)->xfwfFormattedText.ypos = 0;	/* Bottom of previous line */
    ((XfwfFormattedTextWidget) self)->xfwfFormattedText.vspace = 0;
    ((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftypos = 0;	/* Bottom of left floats */
    ((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftvspace = 0;
    ((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightypos = 0;	/* Bottom of right floats */
    ((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightvspace = 0;
    ((XfwfFormattedTextWidget) self)->xfwfFormattedText.next_to_render = NULL;
    ((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrlines = 0;
    ((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrleftlines = 0;
    ((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrrightlines = 0;
}

static void append(Widget self, FtChunk p)
{
    p->next = NULL;
    p->prev = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.last;
    if (!((XfwfFormattedTextWidget) self)->xfwfFormattedText.chunks)
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.chunks = p;
    else
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.last->next = p;
    ((XfwfFormattedTextWidget) self)->xfwfFormattedText.last = p;
}

static String strndup(const String s, Cardinal n)
{
    String t = XtMalloc((n + 1) * sizeof(*t));
    *t = '\0';
    strncat(t, s, n);
    return t;
}

static int find_line_below_y(FtLine * lines, int nrlines, int y)
{
    int l = 0, h = nrlines, m;

    while (l != h) {
	m = (l + h) / 2;
	if (lines[m].y <= y)
	    l = m + 1;
	else
	    h = m;
    }
    assert(l == 0 || lines[l - 1].y <= y);
    return l;
}

static FtChunk local_find_chunk_at_xy(FtLine * lines, int nrlines, int x, int y)
{
    int n;
    FtChunk p;

    n = find_line_below_y(lines, nrlines, y) - 1;
    if (n >= 0 && lines[n].y + lines[n].h > y
	&& lines[n].x <= x && lines[n].x + lines[n].w > x) {
	assert(lines[n].y <= y);
	for (p = lines[n].start; p && p != lines[n].stop; p = p->next)
	    if (p->x <= x && x < p->x + p->w
		&& p->y - p->h <= y && y < p->y + p->d)
		return p;
    }
    return NULL;
}

static FtChunk find_chunk_at_xy(Widget self, int x, int y)
{
    FtChunk p;

    p = local_find_chunk_at_xy(((XfwfFormattedTextWidget) self)->xfwfFormattedText.lines, ((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrlines, x, y);
    if (!p)
	p = local_find_chunk_at_xy(((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftlines, ((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrleftlines, x, y);
    if (!p)
	p = local_find_chunk_at_xy(((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightlines, ((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrrightlines, x, y);
    return p;
}

static void resort_lines_vertically(FtLine * lines, int nrlines)
{
    int i, j;
    FtLine h;

    j = nrlines - 1;		/* The line to insert */
    i = nrlines - 2;
    while (i >= 0 && lines[i].y > lines[j].y)
	i--;
    if (i < nrlines - 2) {
	i++;
	h = lines[j];
	for (; j > i; j--)
	    lines[j] = lines[j - 1];	/* Shift lines */
	lines[i] = h;
    }
}

static void append_line(Widget self, int mode, int x, int y, int w, int h, FtChunk start, FtChunk stop)
{
    if ((mode & FtLEFTFLOAT)) {
	assert(((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrleftlines <= ((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrleftlines_alloc);
	if (((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrleftlines == ((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrleftlines_alloc) {
	    ((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrleftlines_alloc = (((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrleftlines_alloc + INCR) / INCR * INCR;
	    renewarray(((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftlines, ((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrleftlines_alloc);
	}
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftlines[((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrleftlines].x = x;
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftlines[((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrleftlines].y = y;
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftlines[((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrleftlines].w = w;
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftlines[((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrleftlines].h = h;
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftlines[((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrleftlines].start = start;
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftlines[((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrleftlines].stop = stop;
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrleftlines++;
	resort_lines_vertically(((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftlines, ((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrleftlines);
    } else if ((mode & FtRIGHTFLOAT)) {
	assert(((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrrightlines <= ((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrrightlines_alloc);
	if (((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrrightlines == ((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrrightlines_alloc) {
	    ((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrrightlines_alloc = (((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrrightlines_alloc + INCR) / INCR * INCR;
	    renewarray(((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightlines, ((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrrightlines_alloc);
	}
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightlines[((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrrightlines].x = x;
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightlines[((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrrightlines].y = y;
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightlines[((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrrightlines].w = w;
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightlines[((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrrightlines].h = h;
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightlines[((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrrightlines].start = start;
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightlines[((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrrightlines].stop = stop;
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrrightlines++;
	resort_lines_vertically(((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightlines, ((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrrightlines);
    } else {
	assert(((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrlines <= ((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrlines_alloc);
	if (((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrlines == ((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrlines_alloc) {
	    ((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrlines_alloc = (((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrlines_alloc + INCR) / INCR * INCR;
	    renewarray(((XfwfFormattedTextWidget) self)->xfwfFormattedText.lines, ((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrlines_alloc);
	}
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.lines[((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrlines].x = x;
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.lines[((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrlines].y = y;
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.lines[((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrlines].w = w;
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.lines[((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrlines].h = h;
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.lines[((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrlines].start = start;
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.lines[((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrlines].stop = stop;
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrlines++;
	resort_lines_vertically(((XfwfFormattedTextWidget) self)->xfwfFormattedText.lines, ((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrlines);
    }
}

static FtChunk find_parshape_before(FtChunk h)
{
#if 0
    if (h == NULL)
	return NULL;
    else if (h->tp == FtParShape)
	return h;
    else
	return find_parshape_before(h->prev);
#else
    while (h && h->tp != FtParShape)
	h = h->prev;
    return h;
#endif
}

static void render_one_line(Widget self, FtChunk first, FtChunk stop, int mode, int x, int top, int maxht, int maxdp, int targetwd, int sumwd, int sumstretch, Bool may_fill)
{
    FtChunk p;
    int desiredwhite, minwhite, startx, base;

    assert(!is_breakable(first));
    while (stop->prev != first
	   && stop->prev->tp != FtWord && stop->prev->tp != FtInline)
	stop = stop->prev;

    base = top + maxht;
    minwhite = sumstretch * SHRINK;
    if (sumwd + minwhite > targetwd)	/* Doesn't fit any way */
	desiredwhite = minwhite;
    else if (sumwd + sumstretch > targetwd)	/* Needs some shrink */
	desiredwhite = targetwd - sumwd;
    else if ((mode & FtJUSTIFYMASK) == FtJUSTIFYBOTH && may_fill)
	desiredwhite = targetwd - sumwd;
    else			/* Fits with normal spaces */
	desiredwhite = sumstretch;

    if ((mode & FtJUSTIFYMASK) == FtJUSTIFYCENTER)
	x += (targetwd - sumwd - desiredwhite) / 2;
    else if ((mode & FtJUSTIFYMASK) == FtJUSTIFYRIGHT)
	x += targetwd - sumwd - desiredwhite;
    startx = x;

    for (p = first; p != stop; p = p->next) {	/* Set all coordinates */
	p->x = x;
	p->y = base;
	if (p->tp != FtHSpace || !p->u.hspace.stretchable) {
	    x += p->w;
	} else {
	    assert(sumstretch != 0);
	    p->w = p->u.hspace.normwidth * desiredwhite / sumstretch;
	    x += p->w;
	    desiredwhite -= p->w;
	    sumstretch -= p->u.hspace.normwidth;
	}
    }
#if 0
    append_line(self, mode, startx, top, x - startx, maxht + maxdp, first, stop);
#endif
}

static FtChunk find_break(FtChunk h, FtChunk last, int *wd, int *stretch, int *ht, int *dp)
{
    assert(h != last && is_breakable(last));
    *stretch = 0;
    *wd = 0;
    *ht = 0;
    *dp = 0;
    do {
	if (h->tp != FtHSpace || !h->u.hspace.stretchable)
	    *wd += h->w;
	else
	    *stretch += h->u.hspace.normwidth;
	set_max(*ht, h->h);
	set_max(*dp, h->d);
	h = h->next;
    } while (!is_breakable(h));
    return h;
}

static int left_float_width(Widget self, int topy, int boty)
{
    int i, w = 0;
    for (i = 0; i < ((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrleftlines; i++) {
	if (((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftlines[i].y >= boty)
	    break;
	else if (((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftlines[i].y + ((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftlines[i].h < topy)
	    continue;
	else
	    set_max(w, ((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftlines[i].x + ((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftlines[i].w);
    }
    return w;
}

static int right_float_width(Widget self, int topy, int boty)
{
    int i, w = 0;
    for (i = 0; i < ((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrrightlines; i++) {
	if (((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightlines[i].y >= boty)
	    break;
	else if (((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightlines[i].y + ((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightlines[i].h < topy)
	    continue;
	else
	    set_max(w, ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fullwidth - ((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightlines[i].x);
    }
    return w;
}

static int compute_width(Widget self, FtChunk p, int ytop, int ybottom)
{
    int lft, rgt;

    assert(!p || p->tp == FtParShape);
    if (p == NULL) {
	/* No parshape, use full width minus floats */
	lft = left_float_width(self, ytop, ybottom);
	rgt = right_float_width(self, ytop, ybottom);
	return ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fullwidth - lft - rgt;
    } else if ((p->u.parshape.mode & (FtLEFTFLOAT | FtRIGHTFLOAT)) == 0) {
	/* This is normal paragraph */
	lft = left_float_width(self, ytop, ybottom);
	set_max(lft, p->u.parshape.left);
	rgt = right_float_width(self, ytop, ybottom);
	set_max(rgt, p->u.parshape.right);
	return ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fullwidth - lft - rgt;
    } else if (p->u.parshape.width > 0) {
	/* This is a floating par. with explicit width */
	return p->u.parshape.width;
    } else if (p->next && !is_breakable(p->next)) {
	/* This is a floating par. without explicit width */
#if 0
	fprintf(stderr, "float width taken from next chunk (tp=%d): %d\n",
		p->next->tp, p->next->w);
#endif
	return p->next->w;
    } else {
	/* This is floating par. with nothing fixed following it */
#if 0
	fprintf(stderr, "float width estimated: %d\n",
		((XfwfFormattedTextWidget) self)->xfwfFormattedText.fullwidth / 3 - p->u.parshape.left -
		p->u.parshape.right);
#endif
	return ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fullwidth / 3 - p->u.parshape.left - p->u.parshape.right;
    }
}

static FtChunk render_paragraphs(Widget self, FtChunk last)
{
    FtChunk p, p2, q, breakpoint;
    int mode, sumwd, maxht, maxdp, sumstretch;
    int wd, ht, dp, stretchwd, targetwd, targetwd1;
    int maxht1, maxdp1, sumwd1, sumstretch1, y, x, x1, x2;
    float leading;

    assert(last->tp != FtWord && last->tp != FtHSpace && last->tp != FtInline);

    if (!((XfwfFormattedTextWidget) self)->xfwfFormattedText.next_to_render)
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.next_to_render = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.chunks;
    while (((XfwfFormattedTextWidget) self)->xfwfFormattedText.next_to_render != last && is_breakable(((XfwfFormattedTextWidget) self)->xfwfFormattedText.next_to_render))
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.next_to_render = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.next_to_render->next;

    /* Find first parshape to use */
    p = find_parshape_before(((XfwfFormattedTextWidget) self)->xfwfFormattedText.next_to_render);

    while (((XfwfFormattedTextWidget) self)->xfwfFormattedText.next_to_render != last) {
	breakpoint = find_break(((XfwfFormattedTextWidget) self)->xfwfFormattedText.next_to_render, last,
				&sumwd, &sumstretch, &maxht, &maxdp);
	/* sumwd doesn't include the stretchable space */

	assert(!p || p->tp == FtParShape);
	mode = p ? p->u.parshape.mode : 0;
	leading = p ? p->u.parshape.leading : 1.0;

	if (mode & FtRIGHTFLOAT)
	    y = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightypos + ((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightvspace;
	else if (mode & FtLEFTFLOAT)
	    y = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftypos + ((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftvspace;
	else
	    y = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.ypos + ((XfwfFormattedTextWidget) self)->xfwfFormattedText.vspace;

	targetwd = compute_width(self, p, y, y + maxht + maxdp);
	if (targetwd < ((XfwfFormattedTextWidget) self)->core.width / 10) {	/* Try to flush floats */
	    set_max(y, min(((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftypos + ((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftvspace,
			   ((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightypos + ((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightvspace));
	    targetwd = compute_width(self, p, y, y + maxht + maxdp);
	}
	if (targetwd < ((XfwfFormattedTextWidget) self)->core.width / 10) {	/* Flush all floats */
	    set_max(y, max(((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftypos + ((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftvspace,
			   ((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightypos + ((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightvspace));
	    targetwd = compute_width(self, p, y, y + maxht + maxdp);
	}
	while (breakpoint != last) {
	    q = find_break(breakpoint, last, &wd, &stretchwd, &ht, &dp);
	    sumwd1 = sumwd + wd;
	    sumstretch1 = sumstretch + stretchwd;
	    if (sumwd1 + sumstretch1 * SHRINK > targetwd)
		break;		/* Break at last breakpoint */
	    if (ht > maxht || dp > maxdp) {	/* Recompute targetwd */
		maxht1 = max(maxht, ht);
		maxdp1 = max(maxdp, dp);
		targetwd1 = compute_width(self, p, y, y + maxht1 + maxdp1);
		if (sumwd1 + sumstretch1 * SHRINK > targetwd1)
		    break;	/* Doesn't fit, due to ht/dp */
		/* Still fits, despite larger height or depth */
		maxht = maxht1;
		maxdp = maxdp1;
		targetwd = targetwd1;
	    }
	    sumwd = sumwd1;
	    sumstretch = sumstretch1;
	    breakpoint = q;
	}

	/* breakpoint is best breakpoint, render one line */
	assert(is_breakable(breakpoint));
	if (mode & FtLEFTFLOAT) {
	    x1 = 0;
	    x = p->u.parshape.left;
	    x2 = x + targetwd + p->u.parshape.right;
	} else if (mode & FtRIGHTFLOAT) {
	    x = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fullwidth - targetwd - p->u.parshape.right;
	    x1 = x - p->u.parshape.left;
	    x2 = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fullwidth;
	} else if (!p) {
	    x = left_float_width(self, y, y + maxht + maxdp);
	    x1 = x;
	    x2 = x + targetwd;
	} else {
	    x1 = left_float_width(self, y, y + maxht + maxdp);
	    x = max(x1, p->u.parshape.left);
	    x2 = x + targetwd + p->u.parshape.right;
	    /* x2 may be too large, should we compare it to right_float? */
	}
	render_one_line(self, ((XfwfFormattedTextWidget) self)->xfwfFormattedText.next_to_render, breakpoint, mode, x, y, maxht,
			maxdp, targetwd, sumwd, sumstretch,
			breakpoint != last);
#if 1
	append_line(self, mode, x1, y, x2 - x1, maxht + maxdp,
		    ((XfwfFormattedTextWidget) self)->xfwfFormattedText.next_to_render, breakpoint);
#endif

	/* Set variables for next cycle */
	if (mode & FtLEFTFLOAT) {
	    ((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftypos = y + round((maxht + maxdp) * leading);
	    ((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftvspace = 0;
	} else if (mode & FtRIGHTFLOAT) {
	    ((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightypos = y + round((maxht + maxdp) * leading);
	    ((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightvspace = 0;
	} else {
	    ((XfwfFormattedTextWidget) self)->xfwfFormattedText.ypos = y + round((maxht + maxdp) * leading);
	    ((XfwfFormattedTextWidget) self)->xfwfFormattedText.vspace = 0;
	}
	/* Small optimization: if we encounter a parshape,
	 * we don't have to call find_parshape_before() below
	 */
	p2 = NULL;
	while (breakpoint != last && is_breakable(breakpoint)) {
	    if (breakpoint->tp == FtParShape)
		p2 = breakpoint;
	    breakpoint = breakpoint->next;
	}
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.next_to_render = breakpoint;

	/* Find parshape to use for next line */
	p = p2 ? p2 : find_parshape_before(((XfwfFormattedTextWidget) self)->xfwfFormattedText.next_to_render);
    }
    return p;
}

static void format_vspace(Widget self, FtChunk h)
{
    FtChunk p;
    int mode;
    p = render_paragraphs(self, h);
    /* p = find_parshape_before(h); */
    mode = p ? p->u.parshape.mode : FtNORMALPARA;
    switch (mode & FtPOSITIONMASK) {
    case FtNORMALPARA:
	set_max(((XfwfFormattedTextWidget) self)->xfwfFormattedText.vspace, h->h);
	break;
    case FtLEFTFLOAT:
	set_max(((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftvspace, h->h);
	break;
    case FtRIGHTFLOAT:
	set_max(((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightvspace, h->h);
	break;
    default:
	assert(!"Cannot happen");
    }
}

static void format_inline(Widget self, FtChunk h)
{
    Widget w = h->u.inlined.widget;
    int wd, ht, dp, fixedht;
    int vm = h->u.inlined.vmargin, hm = h->u.inlined.hmargin;

    assert(!(h->u.inlined.style & FtHIDEWIDTH) || h->w == 0);
    if ((h->u.inlined.style & (FtHIDEWIDTH | FtALIGNFIXEDWIDTH)) == 0)
	h->w = ((XfwfFormattedTextWidget) w)->core.width + 2 * hm;
    font_space(self, h->u.inlined.style, &ht, &dp, &wd, &h->raise);
    if (h->u.inlined.style & FtALIGNFIXEDHEIGHT) {
	fixedht = h->h + h->d;
	switch (h->u.inlined.style & FtALIGNMASK) {
	case FtALIGNTOP:
	    h->h = ht;
	    h->d = fixedht - ht;
	    break;
	case FtALIGNBOTTOM:
	    h->d = dp;
	    h->h = fixedht - dp;
	    break;
	case FtALIGNMIDDLE:
	    h->h = (fixedht + ht - dp) / 2;
	    h->d = fixedht - h->h;
	    break;
	case FtALIGNDEPTH:
	    h->h = fixedht - h->d;
	    break;
	default:
	    assert(!"Cannot happen");
	}
	assert(h->h + h->d == fixedht);
    } else {			/* No fixed height */
	switch (h->u.inlined.style & FtALIGNMASK) {
	case FtALIGNTOP:
	    h->h = ht + vm;
	    h->d = ((XfwfFormattedTextWidget) w)->core.height - ht + vm;
	    break;
	case FtALIGNBOTTOM:
	    h->d = dp + vm;
	    h->h = ((XfwfFormattedTextWidget) w)->core.height - dp + vm;
	    break;
	case FtALIGNMIDDLE:
	    h->h = (((XfwfFormattedTextWidget) w)->core.height + ht - dp) / 2 + vm;
	    h->d = ((XfwfFormattedTextWidget) w)->core.height - h->h + 2 * vm;
	    break;
	case FtALIGNDEPTH:
	    h->h = ((XfwfFormattedTextWidget) w)->core.height + 2 * vm - h->d;
	    break;
	default:
	    assert(!"Cannot happen");
	}
    }
}

static int find_width_before(Widget self, FtChunk h)
{
    h = h->prev;
    for (; h && h->tp != FtParShape && h->tp != FtInline; h = h->prev);
    if (!h)
	return -1;
    else if (h->tp == FtInline)
	return h->w;
    else if (h->u.parshape.mode & (FtLEFTFLOAT | FtRIGHTFLOAT)) {
	assert(h->u.parshape.width >= 0);
	return h->u.parshape.width;
    } else
	return -1;
}

static void format_parshape(Widget self, FtChunk p)
{
    int width;

    (void) render_paragraphs(self, p);
    switch (p->u.parshape.mode & FtPOSITIONMASK) {
    case FtNORMALPARA:
	if (p->u.parshape.mode & FtCLEARLEFT)
	    set_max(((XfwfFormattedTextWidget) self)->xfwfFormattedText.ypos, ((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftypos + ((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftvspace - ((XfwfFormattedTextWidget) self)->xfwfFormattedText.vspace);
	if (p->u.parshape.mode & FtCLEARRIGHT)
	    set_max(((XfwfFormattedTextWidget) self)->xfwfFormattedText.ypos, ((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightypos + ((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightvspace - ((XfwfFormattedTextWidget) self)->xfwfFormattedText.vspace);
#if 0
	assert(left_float_width(self, ((XfwfFormattedTextWidget) self)->xfwfFormattedText.ypos + ((XfwfFormattedTextWidget) self)->xfwfFormattedText.vspace, ((XfwfFormattedTextWidget) self)->xfwfFormattedText.ypos + ((XfwfFormattedTextWidget) self)->xfwfFormattedText.vspace)
	       + right_float_width(self, ((XfwfFormattedTextWidget) self)->xfwfFormattedText.ypos + ((XfwfFormattedTextWidget) self)->xfwfFormattedText.vspace, ((XfwfFormattedTextWidget) self)->xfwfFormattedText.ypos + ((XfwfFormattedTextWidget) self)->xfwfFormattedText.vspace)
	<= ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fullwidth);
#endif
	break;
    case FtLEFTFLOAT:
	set_max(((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftypos, ((XfwfFormattedTextWidget) self)->xfwfFormattedText.ypos + ((XfwfFormattedTextWidget) self)->xfwfFormattedText.vspace - ((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftvspace);
	if (p->u.parshape.mode & FtCLEARRIGHT)
	    set_max(((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftypos, ((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightypos + ((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightvspace - ((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftvspace);
	if (p->u.parshape.width >= 0);
	else if ((p->u.parshape.mode & FtKEEPWIDTH)
		 && (width = find_width_before(self, p)) >= 0)
	    p->u.parshape.width = width;
#if 0
	else			/* if (p->u.parshape.width < 0) */
	    p->u.parshape.width = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fullwidth / 3
		- p->u.parshape.left - p->u.parshape.right;
#endif
	/* check if there is no collision of floats */
	if (left_float_width(self, ((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftypos + ((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftvspace,
			     ((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftypos + ((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftvspace)
	    + right_float_width(self, ((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftypos + ((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftvspace,
				((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftypos + ((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftvspace) >= ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fullwidth)
	    ((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftypos = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightypos - ((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftvspace;
	break;
    case FtRIGHTFLOAT:
	set_max(((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightypos, ((XfwfFormattedTextWidget) self)->xfwfFormattedText.ypos + ((XfwfFormattedTextWidget) self)->xfwfFormattedText.vspace - ((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightvspace);
	if (p->u.parshape.mode & FtCLEARLEFT)
	    set_max(((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightypos, ((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftypos + ((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftvspace - ((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightvspace);
	if (p->u.parshape.width >= 0);
	else if ((p->u.parshape.mode & FtKEEPWIDTH)
		 && (width = find_width_before(self, p)) >= 0)
	    p->u.parshape.width = width;
#if 0
	else			/* if (p->u.parshape.width < 0) */
	    p->u.parshape.width = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fullwidth / 3
		- p->u.parshape.left - p->u.parshape.right;
#endif
	/* check if there is no collision of floats */
	if (left_float_width(self, ((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightypos + ((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightvspace,
			     ((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightypos + ((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightvspace)
	    + right_float_width(self, ((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightypos + ((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightvspace,
				((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightypos + ((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightvspace) >= ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fullwidth)
	    ((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightypos = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftypos - ((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightvspace;
	break;
    default:
	assert(!"Cannot happen");
    }
}

static void format_hrule(Widget self, FtChunk h)
{
    FtChunk p;
    int mode, left, right;

    assert(h->d >= 0);
    p = render_paragraphs(self, h);
    /* p = find_parshape_before(h); */
    mode = p ? p->u.parshape.mode : FtNORMALPARA;
    left = p ? p->u.parshape.left : 0;
    right = p ? p->u.parshape.right : 0;
    switch (mode & FtPOSITIONMASK) {
    case FtNORMALPARA:
	h->y = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.ypos + ((XfwfFormattedTextWidget) self)->xfwfFormattedText.vspace;
	h->x = left_float_width(self, h->y, h->y + h->d) + left;
	h->w = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fullwidth - h->x - right
	    - right_float_width(self, h->y, h->y + h->d);
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.vspace = 0;
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.ypos = h->y + h->d;
	break;
    case FtLEFTFLOAT:
	h->y = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftypos + ((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftvspace;
	h->x = left;
	assert(p && p->u.parshape.width >= 0);
	h->w = p->u.parshape.width;
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftvspace = 0;
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftypos = h->y + h->d;
	break;
    case FtRIGHTFLOAT:
	h->y = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightypos + ((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightvspace;
	assert(p && p->u.parshape.width >= 0);
	h->w = p->u.parshape.width;
	h->x = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fullwidth - p->u.parshape.width - right;
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightvspace = 0;
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightypos = h->y + h->d;
	break;
    default:
	assert(!"Cannot happen");
    }
    append_line(self, mode, h->x, h->y, h->w, h->d, h, h->next);
}

static void unmap_children(Widget self)
{
    FtChunk p;

    for (p = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.inlined_chunks; p; p = p->u.inlined.prev_inlined) {
	assert(p->tp == FtInline);
#if 0
	if (p->y + p->h - ((XfwfFormattedTextWidget) self)->xfwfFormattedText.voffset > 0 && p->y - ((XfwfFormattedTextWidget) self)->xfwfFormattedText.voffset <= ((XfwfFormattedTextWidget) self)->core.height)
	    XtMoveWidget(p->u.inlined.widget, p->x + p->u.inlined.hmargin,
			 p->y - p->h + p->u.inlined.vmargin - ((XfwfFormattedTextWidget) self)->xfwfFormattedText.voffset);
	else if (p->u.inlined.mapped) {
	    XtUnmapWidget(p->u.inlined.widget);
	    p->u.inlined.mapped = FALSE;
	}
#else
	XtMoveWidget(p->u.inlined.widget, p->x + p->u.inlined.hmargin,
		     p->y - p->h + p->u.inlined.vmargin - ((XfwfFormattedTextWidget) self)->xfwfFormattedText.voffset);
#endif
    }
}

static void scroll_cb(Widget vscroll, XtPointer client_data, XtPointer call_data)
{
    XmScrollBarCallbackStruct *info = (XmScrollBarCallbackStruct *) call_data;
    Widget self = (Widget) client_data;
    int delta;

    delta = info->value - ((XfwfFormattedTextWidget) self)->xfwfFormattedText.voffset;
    ((XfwfFormattedTextWidget) self)->xfwfFormattedText.voffset = info->value;
    unmap_children(self);
#if 1
    XCopyArea(XtDisplay(self), XtWindow(self), XtWindow(self),
	      DefaultGCOfScreen(XtScreen(self)), 0, delta,
	      ((XfwfFormattedTextWidget) self)->core.width - ((XfwfFormattedTextWidget) vscroll)->core.width, ((XfwfFormattedTextWidget) self)->core.height, 0, 0);
#else
    XClearArea(XtDisplay(self), XtWindow(self), 0, 0, 0, 0, TRUE);
#endif
}

static Boolean do_layout(XtPointer client_data)
{
    Widget self = (Widget) client_data;
    FtChunk p;

    init_par_builder(self);
    for (p = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.chunks; p; p = p->next) {
	switch (p->tp) {
	case FtWord:
	    break;
	case FtHSpace:
	    break;
	case FtInline:
	    format_inline(self, p);
	    break;
	case FtVSpace:
	    format_vspace(self, p);
	    break;
	case FtParShape:
	    format_parshape(self, p);
	    break;
	case FtHRule:
	    format_hrule(self, p);
	    break;
	default:
	    assert(!"Cannot happen");
	}
    }
    unmap_children(self);
    if (XtIsRealized(self)) {
	XClearArea(XtDisplay(self), XtWindow(self), 0, 0, 0, 0, TRUE);
	/* $expose($, NULL, NULL); */
    }
    ((XfwfFormattedTextWidgetClass) self->core.widget_class)->xfwfFormattedText_class.reformat_scrollbar(self);
    ((XfwfFormattedTextWidget) self)->xfwfFormattedText.workid = 0;
    return TRUE;		/* Un-install */
}


static void install_do_layout(Widget self)
{
    if (!((XfwfFormattedTextWidget) self)->xfwfFormattedText.workid)
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.workid = XtAppAddWorkProc(XtWidgetToApplicationContext(self),
							do_layout, self);
}


static void load_font(Widget self, int family, int style, int size)
{
    unsigned long val;
    char s[256], *fam, *charset, *slant, *bold;
    XFontStruct *fs;
    int siz, hres, vres;

    assert(0 <= family && family < NR_FAMILIES);
    switch (family) {
    case 4:
	fam = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fontFamily5;
	charset = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fontCharset5;
	slant = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fontSlant5;
	bold = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fontBold5;
	break;
    case 3:
	fam = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fontFamily4;
	charset = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fontCharset4;
	slant = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fontSlant4;
	bold = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fontBold4;
	break;
    case 2:
	fam = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fontFamily3;
	charset = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fontCharset3;
	slant = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fontSlant3;
	bold = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fontBold3;
	break;
    case 1:
	fam = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fontFamily2;
	charset = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fontCharset2;
	slant = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fontSlant2;
	bold = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fontBold2;
	break;
    default:
	fam = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fontFamily1;
	charset = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fontCharset1;
	slant = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fontSlant1;
	bold = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fontBold1;
	break;
    }
    assert(0 <= size && size < NR_SIZES);
    siz = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[family][size];
    hres = 25.4 * WidthOfScreen(XtScreen(self)) / WidthMMOfScreen(XtScreen(self));
    vres = 25.4 * HeightOfScreen(XtScreen(self)) / HeightMMOfScreen(XtScreen(self));
    if (abs(hres - 75) > abs(hres - 100))
	hres = 100;
    else
	hres = 75;
    if (abs(vres - 75) > abs(vres - 100))
	vres = 100;
    else
	vres = 75;

    switch (style) {
    case 0:			/* Normal style */
	sprintf(s, "-*-%s-medium-r-normal--*-%d-%d-%d-*-*-%s",
		fam, siz, hres, vres, charset);
	fs = XLoadQueryFont(XtDisplay(self), s);
	if (fs)
	    break;
	sprintf(s, "-*-%s-medium-r-*--*-%d-*-*-*-*-%s", fam, siz, charset);
	fs = XLoadQueryFont(XtDisplay(self), s);
	if (fs)
	    break;
	sprintf(s, "-*-*-medium-r-*--*-%d-*-*-*-*-%s", siz, charset);
	fs = XLoadQueryFont(XtDisplay(self), s);
	if (fs)
	    break;
	sprintf(s, "fixed");
	fs = XLoadQueryFont(XtDisplay(self), s);
	if (fs)
	    break;
	XtAppError(XtWidgetToApplicationContext(self), "Failed to find Xfonts");
	break;
    case 1:			/* Bold style */
	sprintf(s, "-*-%s-%s-r-normal--*-%d-%d-%d-*-*-%s",
		fam, bold, siz, hres, vres, charset);
	fs = XLoadQueryFont(XtDisplay(self), s);
	if (fs)
	    break;
	sprintf(s, "-*-%s-%s-r-*--*-%d-*-*-*-*-%s",
		fam, bold, siz, charset);
	fs = XLoadQueryFont(XtDisplay(self), s);
	if (fs)
	    break;
	sprintf(s, "-*-%s-*-r-*--*-%d-*-*-*-*-%s",
		fam, siz, charset);
	fs = XLoadQueryFont(XtDisplay(self), s);
	if (fs)
	    break;
	sprintf(s, "fixed");
	fs = XLoadQueryFont(XtDisplay(self), s);
	if (fs)
	    break;
	XtAppError(XtWidgetToApplicationContext(self), "Failed to find Xfonts");
	break;
    case 2:			/* Italic style */
	sprintf(s, "-*-%s-medium-%s-normal--*-%d-%d-%d-*-*-%s",
		fam, slant, siz, hres, vres, charset);
	fs = XLoadQueryFont(XtDisplay(self), s);
	if (fs)
	    break;
	sprintf(s, "-*-%s-medium-%s-*--*-%d-*-*-*-*-%s",
		fam, slant, siz, charset);
	fs = XLoadQueryFont(XtDisplay(self), s);
	if (fs)
	    break;
	sprintf(s, "-*-%s-*-%s-*--*-%d-*-*-*-*-%s",
		fam, slant, siz, charset);
	fs = XLoadQueryFont(XtDisplay(self), s);
	if (fs)
	    break;
	sprintf(s, "fixed");
	fs = XLoadQueryFont(XtDisplay(self), s);
	if (fs)
	    break;
	XtAppError(XtWidgetToApplicationContext(self), "Failed to find Xfonts");
	break;
    case 3:			/* Bold-italic style */
	sprintf(s, "-*-%s-%s-%s-normal--*-%d-%d-%d-*-*-%s",
		fam, bold, slant, siz, hres, vres, charset);
	fs = XLoadQueryFont(XtDisplay(self), s);
	if (fs)
	    break;
	sprintf(s, "-*-%s-%s-%s-*--*-%d-*-*-*-*-%s",
		fam, bold, slant, siz, charset);
	fs = XLoadQueryFont(XtDisplay(self), s);
	if (fs)
	    break;
	sprintf(s, "-*-%s-*-%s-*--*-%d-*-*-*-*-%s",
		fam, slant, siz, charset);
	fs = XLoadQueryFont(XtDisplay(self), s);
	if (fs)
	    break;
	sprintf(s, "-*-%s-%s-*-*--*-%d-*-*-*-*-%s",
		fam, bold, siz, charset);
	fs = XLoadQueryFont(XtDisplay(self), s);
	if (fs)
	    break;
	sprintf(s, "fixed");
	fs = XLoadQueryFont(XtDisplay(self), s);
	if (fs)
	    break;
	XtAppError(XtWidgetToApplicationContext(self), "Failed to find Xfonts");
	break;
    default:
	assert(!"Cannot happen");
    }
    if (XGetFontProperty(fs, XA_SUPERSCRIPT_Y, &val))
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.fonts[family][style][size].sup = val;
    else
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.fonts[family][style][size].sup = fs->ascent / 2;
    if (XGetFontProperty(fs, XA_SUBSCRIPT_Y, &val))
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.fonts[family][style][size].sub = val;
    else
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.fonts[family][style][size].sub = -fs->descent / 2;
    if (XGetFontProperty(fs, XA_UNDERLINE_POSITION, &val))
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.fonts[family][style][size].ulpos = val;
    else
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.fonts[family][style][size].ulpos = 2;
    if (XGetFontProperty(fs, XA_UNDERLINE_THICKNESS, &val))
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.fonts[family][style][size].ulthick = val;
    else
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.fonts[family][style][size].ulthick = 1;
    if (XGetFontProperty(fs, XA_NORM_SPACE, &val))
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.fonts[family][style][size].space = val;
    else
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.fonts[family][style][size].space = XTextWidth(fs, " ", 1);
    if (XGetFontProperty(fs, XA_QUAD_WIDTH, &val))
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.fonts[family][style][size].em = val;
    else
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.fonts[family][style][size].em = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fonts[family][style][size].space;
    ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fonts[family][style][size].fs = fs;
}

static void interpret_style(Widget self, TextStyle style, int *fam, int *sty, int *siz, int *raise)
{
    int sup, sub;

    *fam = (style & FtFAMILYMASK) >> FtFAMILYSHIFT;
    *sty = (style & FtSTYLEMASK) >> FtSTYLESHIFT;
    *siz = (style & FtSIZEMASK) >> FtSIZESHIFT;
    sup = (style & FtSUPERMASK) >> FtSUPERSHIFT;
    sub = (style & FtSUBMASK) >> FtSUBSHIFT;
    if (!((XfwfFormattedTextWidget) self)->xfwfFormattedText.fonts[*fam][*sty][*siz].fs)
	load_font(self, *fam, *sty, *siz);
    *raise = 0;
    for (; sup != 0; sup--) {
	*raise += ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fonts[*fam][*sty][*siz].sup;
	if (*siz != 0)
	    *siz--;
	if (!((XfwfFormattedTextWidget) self)->xfwfFormattedText.fonts[*fam][*sty][*siz].fs)
	    load_font(self, *fam, *sty, *siz);
    }
    for (; sub != 0; sub--) {
	*raise += ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fonts[*fam][*sty][*siz].sup;
	if (*siz != 0)
	    *siz--;
	if (!((XfwfFormattedTextWidget) self)->xfwfFormattedText.fonts[*fam][*sty][*siz].fs)
	    load_font(self, *fam, *sty, *siz);
    }
}

static void font_extents(Widget self, TextStyle style, const String word, int len, int *h, int *d, int *w, int *raise)
{
    int fam, sty, siz;
    interpret_style(self, style, &fam, &sty, &siz, raise);
    *w = XTextWidth(((XfwfFormattedTextWidget) self)->xfwfFormattedText.fonts[fam][sty][siz].fs, word, len);
    *h = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fonts[fam][sty][siz].fs->ascent + *raise;
    *d = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fonts[fam][sty][siz].fs->descent - *raise;
}

static void font_space(Widget self, TextStyle style, int *h, int *d, int *w, int *raise)
{
    int fam, sty, siz;
    interpret_style(self, style, &fam, &sty, &siz, raise);
    *w = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fonts[fam][sty][siz].space;
    *h = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fonts[fam][sty][siz].fs->ascent + *raise;
    *d = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fonts[fam][sty][siz].fs->descent - *raise;
}

static Font font_id(Widget self, TextStyle style)
{
    int fam, sty, siz, raise;
    interpret_style(self, style, &fam, &sty, &siz, &raise);
    return ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fonts[fam][sty][siz].fs->fid;
}

static void dump_chunk(FtChunk f)
{
    switch (f->tp) {
    case FtWord:
	fprintf(stderr, "[FtWord, %d, %d, %d, %d + %d, \"%s\" [%s, %o]]\n",
		f->x, f->y, f->w, f->h, f->d,
		f->data ? f->data : "<none>",
		f->u.word.text, f->u.word.style);
	break;
    case FtHSpace:
	fprintf(stderr, "[FtHSpace, %d, %d, %d, %d + %d, \"%s\" [%d, %o]]\n",
		f->x, f->y, f->w, f->h, f->d,
		f->data ? f->data : "<none>",
		f->u.hspace.normwidth, f->u.hspace.style);
	break;
    case FtInline:
	fprintf(stderr, "[FtInline, %d, %d, %d, %d + %d, \"%s\" [%ld, %o]]\n",
		f->x, f->y, f->w, f->h, f->d,
		f->data ? f->data : "<none>",
		f->u.inlined.widget, f->u.inlined.style);
	break;
    case FtVSpace:
	fprintf(stderr, "[FtVSpace, %d, %d, %d, %d + %d, \"%s\"]\n",
		f->x, f->y, f->w, f->h, f->d,
		f->data ? f->data : "<none>");
	break;
    case FtParShape:
	fprintf(stderr, "[FtParShape [%o, %d, %d, %d]]\n",
		f->u.parshape.mode, f->u.parshape.left,
		f->u.parshape.right, f->u.parshape.width);
	break;
    case FtHRule:
	fprintf(stderr, "[FtHRule, %d, %d, %d, %d + %d, \"%s\"]\n",
		f->x, f->y, f->w, f->h, f->d,
		f->data ? f->data : "<none>");
	break;
    default:
	fprintf(stderr, "???\n");
    }
}

static void dump(Widget self)
{
    FtChunk f;

    for (f = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.chunks; f; f = f->next)
	dump_chunk(f);
}

static XtResource resources[] =
{
    {XtNfontFamily1, XtCFontFamily1, XtRString, sizeof(((XfwfFormattedTextRec *) NULL)->xfwfFormattedText.fontFamily1), XtOffsetOf(XfwfFormattedTextRec, xfwfFormattedText.fontFamily1), XtRImmediate, (XtPointer) "helvetica"},
    {XtNfontCharset1, XtCFontCharset1, XtRString, sizeof(((XfwfFormattedTextRec *) NULL)->xfwfFormattedText.fontCharset1), XtOffsetOf(XfwfFormattedTextRec, xfwfFormattedText.fontCharset1), XtRImmediate, (XtPointer) "iso8859-1"},
    {XtNfontSlant1, XtCFontSlant1, XtRString, sizeof(((XfwfFormattedTextRec *) NULL)->xfwfFormattedText.fontSlant1), XtOffsetOf(XfwfFormattedTextRec, xfwfFormattedText.fontSlant1), XtRImmediate, (XtPointer) "i"},
    {XtNfontBold1, XtCFontBold1, XtRString, sizeof(((XfwfFormattedTextRec *) NULL)->xfwfFormattedText.fontBold1), XtOffsetOf(XfwfFormattedTextRec, xfwfFormattedText.fontBold1), XtRImmediate, (XtPointer) "bold"},
    {XtNfontSizes1, XtCFontSizes1, XtRString, sizeof(((XfwfFormattedTextRec *) NULL)->xfwfFormattedText.fontSizes1), XtOffsetOf(XfwfFormattedTextRec, xfwfFormattedText.fontSizes1), XtRImmediate, (XtPointer) "72 80 100 120 140 180 240"},
    {XtNfontFamily2, XtCFontFamily2, XtRString, sizeof(((XfwfFormattedTextRec *) NULL)->xfwfFormattedText.fontFamily2), XtOffsetOf(XfwfFormattedTextRec, xfwfFormattedText.fontFamily2), XtRImmediate, (XtPointer) "times"},
    {XtNfontCharset2, XtCFontCharset2, XtRString, sizeof(((XfwfFormattedTextRec *) NULL)->xfwfFormattedText.fontCharset2), XtOffsetOf(XfwfFormattedTextRec, xfwfFormattedText.fontCharset2), XtRImmediate, (XtPointer) "iso8859-1"},
    {XtNfontSlant2, XtCFontSlant2, XtRString, sizeof(((XfwfFormattedTextRec *) NULL)->xfwfFormattedText.fontSlant2), XtOffsetOf(XfwfFormattedTextRec, xfwfFormattedText.fontSlant2), XtRImmediate, (XtPointer) "o"},
    {XtNfontBold2, XtCFontBold2, XtRString, sizeof(((XfwfFormattedTextRec *) NULL)->xfwfFormattedText.fontBold2), XtOffsetOf(XfwfFormattedTextRec, xfwfFormattedText.fontBold2), XtRImmediate, (XtPointer) "bold"},
    {XtNfontSizes2, XtCFontSizes2, XtRString, sizeof(((XfwfFormattedTextRec *) NULL)->xfwfFormattedText.fontSizes2), XtOffsetOf(XfwfFormattedTextRec, xfwfFormattedText.fontSizes2), XtRImmediate, (XtPointer) "72 80 100 120 140 180 240"},
    {XtNfontFamily3, XtCFontFamily3, XtRString, sizeof(((XfwfFormattedTextRec *) NULL)->xfwfFormattedText.fontFamily3), XtOffsetOf(XfwfFormattedTextRec, xfwfFormattedText.fontFamily3), XtRImmediate, (XtPointer) "courier"},
    {XtNfontCharset3, XtCFontCharset3, XtRString, sizeof(((XfwfFormattedTextRec *) NULL)->xfwfFormattedText.fontCharset3), XtOffsetOf(XfwfFormattedTextRec, xfwfFormattedText.fontCharset3), XtRImmediate, (XtPointer) "iso8859-1"},
    {XtNfontSlant3, XtCFontSlant3, XtRString, sizeof(((XfwfFormattedTextRec *) NULL)->xfwfFormattedText.fontSlant3), XtOffsetOf(XfwfFormattedTextRec, xfwfFormattedText.fontSlant3), XtRImmediate, (XtPointer) "o"},
    {XtNfontBold3, XtCFontBold3, XtRString, sizeof(((XfwfFormattedTextRec *) NULL)->xfwfFormattedText.fontBold3), XtOffsetOf(XfwfFormattedTextRec, xfwfFormattedText.fontBold3), XtRImmediate, (XtPointer) "bold"},
    {XtNfontSizes3, XtCFontSizes3, XtRString, sizeof(((XfwfFormattedTextRec *) NULL)->xfwfFormattedText.fontSizes3), XtOffsetOf(XfwfFormattedTextRec, xfwfFormattedText.fontSizes3), XtRImmediate, (XtPointer) "72 80 100 120 140 180 240"},
    {XtNfontFamily4, XtCFontFamily4, XtRString, sizeof(((XfwfFormattedTextRec *) NULL)->xfwfFormattedText.fontFamily4), XtOffsetOf(XfwfFormattedTextRec, xfwfFormattedText.fontFamily4), XtRImmediate, (XtPointer) "wwwicons"},
    {XtNfontCharset4, XtCFontCharset4, XtRString, sizeof(((XfwfFormattedTextRec *) NULL)->xfwfFormattedText.fontCharset4), XtOffsetOf(XfwfFormattedTextRec, xfwfFormattedText.fontCharset4), XtRImmediate, (XtPointer) "*"},
    {XtNfontSlant4, XtCFontSlant4, XtRString, sizeof(((XfwfFormattedTextRec *) NULL)->xfwfFormattedText.fontSlant4), XtOffsetOf(XfwfFormattedTextRec, xfwfFormattedText.fontSlant4), XtRImmediate, (XtPointer) "*"},
    {XtNfontBold4, XtCFontBold4, XtRString, sizeof(((XfwfFormattedTextRec *) NULL)->xfwfFormattedText.fontBold4), XtOffsetOf(XfwfFormattedTextRec, xfwfFormattedText.fontBold4), XtRImmediate, (XtPointer) "*"},
    {XtNfontSizes4, XtCFontSizes4, XtRString, sizeof(((XfwfFormattedTextRec *) NULL)->xfwfFormattedText.fontSizes4), XtOffsetOf(XfwfFormattedTextRec, xfwfFormattedText.fontSizes4), XtRImmediate, (XtPointer) "72 80 100 120 140 180 240"},
    {XtNfontFamily5, XtCFontFamily5, XtRString, sizeof(((XfwfFormattedTextRec *) NULL)->xfwfFormattedText.fontFamily5), XtOffsetOf(XfwfFormattedTextRec, xfwfFormattedText.fontFamily5), XtRImmediate, (XtPointer) "symbol"},
    {XtNfontCharset5, XtCFontCharset5, XtRString, sizeof(((XfwfFormattedTextRec *) NULL)->xfwfFormattedText.fontCharset5), XtOffsetOf(XfwfFormattedTextRec, xfwfFormattedText.fontCharset5), XtRImmediate, (XtPointer) "*"},
    {XtNfontSlant5, XtCFontSlant5, XtRString, sizeof(((XfwfFormattedTextRec *) NULL)->xfwfFormattedText.fontSlant5), XtOffsetOf(XfwfFormattedTextRec, xfwfFormattedText.fontSlant5), XtRImmediate, (XtPointer) "*"},
    {XtNfontBold5, XtCFontBold5, XtRString, sizeof(((XfwfFormattedTextRec *) NULL)->xfwfFormattedText.fontBold5), XtOffsetOf(XfwfFormattedTextRec, xfwfFormattedText.fontBold5), XtRImmediate, (XtPointer) "*"},
    {XtNfontSizes5, XtCFontSizes5, XtRString, sizeof(((XfwfFormattedTextRec *) NULL)->xfwfFormattedText.fontSizes5), XtOffsetOf(XfwfFormattedTextRec, xfwfFormattedText.fontSizes5), XtRImmediate, (XtPointer) "72 80 100 120 140 180 240"},
    {XtNactivate, XtCActivate, XtRCallback, sizeof(((XfwfFormattedTextRec *) NULL)->xfwfFormattedText.activate), XtOffsetOf(XfwfFormattedTextRec, xfwfFormattedText.activate), XtRImmediate, (XtPointer) NULL},
    {XtNenter, XtCEnter, XtRCallback, sizeof(((XfwfFormattedTextRec *) NULL)->xfwfFormattedText.enter), XtOffsetOf(XfwfFormattedTextRec, xfwfFormattedText.enter), XtRImmediate, (XtPointer) NULL},
    {XtNleave, XtCLeave, XtRCallback, sizeof(((XfwfFormattedTextRec *) NULL)->xfwfFormattedText.leave), XtOffsetOf(XfwfFormattedTextRec, xfwfFormattedText.leave), XtRImmediate, (XtPointer) NULL},
    {XtNactiveCursor, XtCActiveCursor, XtRCursor, sizeof(((XfwfFormattedTextRec *) NULL)->xfwfFormattedText.activeCursor), XtOffsetOf(XfwfFormattedTextRec, xfwfFormattedText.activeCursor), XtRImmediate, (XtPointer) NULL},
    {XtNwidth, XtCWidth, XtRDimension, sizeof(((XfwfFormattedTextRec *) NULL)->core.width), XtOffsetOf(XfwfFormattedTextRec, core.width), XtRImmediate, (XtPointer) 150},
    {XtNheight, XtCHeight, XtRDimension, sizeof(((XfwfFormattedTextRec *) NULL)->core.height), XtOffsetOf(XfwfFormattedTextRec, core.height), XtRImmediate, (XtPointer) 150},
};

XfwfFormattedTextClassRec xfwfFormattedTextClassRec =
{
    {				/* core_class part */
/* superclass           */ (WidgetClass) & xmManagerClassRec,
/* class_name           */ "XfwfFormattedText",
/* widget_size          */ sizeof(XfwfFormattedTextRec),
/* class_initialize     */ NULL,
/* class_part_initialize */ _resolve_inheritance,
/* class_inited         */ FALSE,
/* initialize           */ initialize,
/* initialize_hook      */ NULL,
/* realize              */ XtInheritRealize,
/* actions              */ actionsList,
/* num_actions          */ 2,
/* resources            */ resources,
/* num_resources        */ 31,
/* xrm_class            */ NULLQUARK,
/* compres_motion       */ FALSE,
/* compress_exposure    */ XtExposeCompressMultiple | XtExposeGraphicsExposeMerged,
/* compress_enterleave  */ True,
/* visible_interest     */ False,
/* destroy              */ destroy,
/* resize               */ resize,
/* expose               */ expose,
/* set_values           */ NULL,
/* set_values_hook      */ NULL,
/* set_values_almost    */ XtInheritSetValuesAlmost,
/* get_values+hook      */ NULL,
/* accept_focus         */ XtInheritAcceptFocus,
/* version              */ XtVersion,
/* callback_private     */ NULL,
/* tm_table             */ defaultTranslations,
/* query_geometry       */ query_geometry,
/* display_acceleator   */ XtInheritDisplayAccelerator,
/* extension            */ NULL
    },
    {				/* composite_class part */
	geometry_manager,
	XtInheritChangeManaged,
	XtInheritInsertChild,
	XtInheritDeleteChild,
	NULL
    },
    {				/* constraint_class part */
/* constraint_resources     */ NULL,
/* num_constraint_resources */ 0,
/* constraint_size          */ sizeof(XfwfFormattedTextConstraintRec),
/* constraint_initialize    */ NULL,
/* constraint_destroy       */ NULL,
/* constraint_set_values    */ NULL,
/* constraint_extension     */ NULL
    },
    {				/* XmManager class part */
#define manager_extension extension
/* translations                 */ XtInheritTranslations,
/* syn_resources                */ NULL,
/* num_syn_resources            */ 0,
/* syn_constraint_resources     */ NULL,
/* num_syn_constraint_resources */ 0,
/* parent_process               */ XmInheritParentProcess,
/* manager_extension            */ NULL,
    },
    {				/* XfwfFormattedText_class part */
	add_word,
	add_hspace,
	add_vspace,
	add_inline,
	add_parshape,
	add_hrule,
	add_eod,
	pass_click,
	reformat_scrollbar,
	get_em_of_textstyle,
	get_space_of_textstyle,
	get_lineheight_of_textstyle,
	get_word_extent,
    },
};
WidgetClass xfwfFormattedTextWidgetClass = (WidgetClass) & xfwfFormattedTextClassRec;
/*ARGSUSED */
static void activate(self, event, params, num_params)
Widget self;
XEvent *event;
String *params;
Cardinal *num_params;
{
    XfwfFTextCallbackStruct info;
    int x, y;
    FtChunk p;

    switch (event->type) {
    case ButtonPress:
    case ButtonRelease:
	x = event->xbutton.x;
	y = event->xbutton.y;
	break;
    case MotionNotify:
	x = event->xmotion.x;
	y = event->xmotion.y;
	break;
    case EnterNotify:
    case LeaveNotify:
	x = event->xcrossing.x;
	y = event->xcrossing.y;
	break;
    default:
	XtAppError(XtWidgetToApplicationContext(self),
		   "activate() action called with incorrect event");
    }
    y += ((XfwfFormattedTextWidget) self)->xfwfFormattedText.voffset;
    /* fprintf(stderr, "activate at (virtual) (%d,%d)\n", x, y); */
    if ((p = find_chunk_at_xy(self, x, y))) {
	if (p->data) {
	    info.reason = XfwfActivate;
	    info.event = event;
	    info.x = x;
	    info.y = y - ((XfwfFormattedTextWidget) self)->xfwfFormattedText.voffset;
	    info.data = p->data;
	    XtCallCallbackList(self, ((XfwfFormattedTextWidget) self)->xfwfFormattedText.activate, &info);
	}
    }
}

/*ARGSUSED */
static void motion(self, event, params, num_params)
Widget self;
XEvent *event;
String *params;
Cardinal *num_params;
{
    XfwfFTextCallbackStruct info;
    int x, y;
    FtChunk found;

    switch (event->type) {
    case ButtonPress:
    case ButtonRelease:
	x = event->xbutton.x;
	y = event->xbutton.y;
	break;
    case MotionNotify:
	x = event->xmotion.x;
	y = event->xmotion.y;
	break;
    case EnterNotify:
    case LeaveNotify:
	x = event->xcrossing.x;
	y = event->xcrossing.y;
	break;
    default:
	XtAppError(XtWidgetToApplicationContext(self),
		   "motion() action called with incorrect event");
    }
    y += ((XfwfFormattedTextWidget) self)->xfwfFormattedText.voffset;
    found = find_chunk_at_xy(self, x, y);
    if (found != ((XfwfFormattedTextWidget) self)->xfwfFormattedText.last_motion_chunk) {
	if (((XfwfFormattedTextWidget) self)->xfwfFormattedText.last_motion_chunk != NULL && ((XfwfFormattedTextWidget) self)->xfwfFormattedText.last_motion_chunk->data != NULL) {
	    info.reason = XfwfLeave;
	    info.event = event;
	    info.x = x;
	    info.y = y - ((XfwfFormattedTextWidget) self)->xfwfFormattedText.voffset;
	    info.data = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.last_motion_chunk->data;
	    XtCallCallbackList(self, ((XfwfFormattedTextWidget) self)->xfwfFormattedText.leave, &info);
	}
	if (found != NULL && found->data != NULL) {
	    XDefineCursor(XtDisplay(self), XtWindow(self), ((XfwfFormattedTextWidget) self)->xfwfFormattedText.activeCursor);
	    info.reason = XfwfEnter;
	    info.event = event;
	    info.x = x;
	    info.y = y - ((XfwfFormattedTextWidget) self)->xfwfFormattedText.voffset;
	    info.data = found->data;
	    XtCallCallbackList(self, ((XfwfFormattedTextWidget) self)->xfwfFormattedText.enter, &info);
	} else {
	    XDefineCursor(XtDisplay(self), XtWindow(self), None);
	}
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.last_motion_chunk = found;
    }
}

static void _resolve_inheritance(class)
WidgetClass class;
{
    XfwfFormattedTextWidgetClass c = (XfwfFormattedTextWidgetClass) class;
    XfwfFormattedTextWidgetClass super;
    static CompositeClassExtensionRec extension_rec =
    {
	NULL, NULLQUARK, XtCompositeExtensionVersion,
	sizeof(CompositeClassExtensionRec), True};
    CompositeClassExtensionRec *ext;
    ext = (XtPointer) XtMalloc(sizeof(*ext));
    *ext = extension_rec;
    ext->next_extension = c->composite_class.extension;
    c->composite_class.extension = ext;
    if (class == xfwfFormattedTextWidgetClass)
	return;
    super = (XfwfFormattedTextWidgetClass) class->core_class.superclass;
    if (c->xfwfFormattedText_class.add_word == XtInherit_add_word)
	c->xfwfFormattedText_class.add_word = super->xfwfFormattedText_class.add_word;
    if (c->xfwfFormattedText_class.add_hspace == XtInherit_add_hspace)
	c->xfwfFormattedText_class.add_hspace = super->xfwfFormattedText_class.add_hspace;
    if (c->xfwfFormattedText_class.add_vspace == XtInherit_add_vspace)
	c->xfwfFormattedText_class.add_vspace = super->xfwfFormattedText_class.add_vspace;
    if (c->xfwfFormattedText_class.add_inline == XtInherit_add_inline)
	c->xfwfFormattedText_class.add_inline = super->xfwfFormattedText_class.add_inline;
    if (c->xfwfFormattedText_class.add_parshape == XtInherit_add_parshape)
	c->xfwfFormattedText_class.add_parshape = super->xfwfFormattedText_class.add_parshape;
    if (c->xfwfFormattedText_class.add_hrule == XtInherit_add_hrule)
	c->xfwfFormattedText_class.add_hrule = super->xfwfFormattedText_class.add_hrule;
    if (c->xfwfFormattedText_class.add_eod == XtInherit_add_eod)
	c->xfwfFormattedText_class.add_eod = super->xfwfFormattedText_class.add_eod;
    if (c->xfwfFormattedText_class.pass_click == XtInherit_pass_click)
	c->xfwfFormattedText_class.pass_click = super->xfwfFormattedText_class.pass_click;
    if (c->xfwfFormattedText_class.reformat_scrollbar == XtInherit_reformat_scrollbar)
	c->xfwfFormattedText_class.reformat_scrollbar = super->xfwfFormattedText_class.reformat_scrollbar;
    if (c->xfwfFormattedText_class.get_em_of_textstyle == XtInherit_get_em_of_textstyle)
	c->xfwfFormattedText_class.get_em_of_textstyle = super->xfwfFormattedText_class.get_em_of_textstyle;
    if (c->xfwfFormattedText_class.get_space_of_textstyle == XtInherit_get_space_of_textstyle)
	c->xfwfFormattedText_class.get_space_of_textstyle = super->xfwfFormattedText_class.get_space_of_textstyle;
    if (c->xfwfFormattedText_class.get_lineheight_of_textstyle == XtInherit_get_lineheight_of_textstyle)
	c->xfwfFormattedText_class.get_lineheight_of_textstyle = super->xfwfFormattedText_class.get_lineheight_of_textstyle;
    if (c->xfwfFormattedText_class.get_word_extent == XtInherit_get_word_extent)
	c->xfwfFormattedText_class.get_word_extent = super->xfwfFormattedText_class.get_word_extent;
}

static void initialize(Widget request, Widget self, ArgList args, Cardinal * num_args)
{
    int i, j, k, n;
    XtGCMask mask, dynmask, dontcaremask;
    XGCValues values;
    Widget vscroll;

    vscroll = XtVaCreateManagedWidget
	("VertScrollBar", xmScrollBarWidgetClass, self,
	 XmNpageIncrement, max(10, ((XfwfFormattedTextWidget) self)->core.height - 10),
	 XmNheight, ((XfwfFormattedTextWidget) self)->core.height,
	 NULL);
    XtMoveWidget(vscroll, ((XfwfFormattedTextWidget) self)->core.width - ((XfwfFormattedTextWidget) vscroll)->core.width, 0);
    XtAddCallback(vscroll, XmNvalueChangedCallback, scroll_cb, self);
    XtAddCallback(vscroll, XmNdragCallback, scroll_cb, self);
    ((XfwfFormattedTextWidget) self)->xfwfFormattedText.vscrollbar = vscroll;
    ((XfwfFormattedTextWidget) self)->xfwfFormattedText.voffset = 0;

    ((XfwfFormattedTextWidget) self)->xfwfFormattedText.last_motion_chunk = NULL;

    mask = 0;
    dynmask = GCForeground | GCBackground | GCFont | GCLineWidth | GCClipMask;
    dontcaremask = GCJoinStyle | GCFillRule | GCTile
	| GCTileStipXOrigin | GCTileStipYOrigin | GCDashOffset | GCDashList
	| GCStipple | GCGraphicsExposures | GCArcMode;
    ((XfwfFormattedTextWidget) self)->xfwfFormattedText.gc = XtAllocateGC(self, ((XfwfFormattedTextWidget) self)->core.depth, mask, &values, dynmask, dontcaremask);

    mask = 0;
    dynmask = GCForeground | GCClipMask;
    dontcaremask = GCBackground | GCFont
	| GCJoinStyle | GCFillRule | GCTile | GCTileStipXOrigin
	| GCTileStipYOrigin | GCDashOffset | GCDashList | GCLineWidth
	| GCGraphicsExposures | GCArcMode | GCLineStyle | GCCapStyle
	| GCStipple;
    ((XfwfFormattedTextWidget) self)->xfwfFormattedText.spacegc = XtAllocateGC(self, ((XfwfFormattedTextWidget) self)->core.depth, mask, &values, dynmask, dontcaremask);

    values.stipple = XCreateBitmapFromData
	(XtDisplay(self), RootWindowOfScreen(XtScreen(self)), line3d_bits,
	 line3d_width, line3d_height);
    values.fill_style = FillOpaqueStippled;
    values.foreground = ((XfwfFormattedTextWidget) self)->xmManager.bottom_shadow_color;
    values.background = ((XfwfFormattedTextWidget) self)->xmManager.top_shadow_color;
    mask = GCFillStyle | GCStipple | GCFillStyle | GCForeground | GCBackground;
    dynmask = GCTileStipXOrigin | GCTileStipYOrigin | GCClipMask;
    dontcaremask = GCLineStyle | GCCapStyle | GCJoinStyle | GCFillRule
	| GCTile | GCDashOffset | GCDashList | GCFont | GCLineWidth
	| GCArcMode | GCGraphicsExposures;
    ((XfwfFormattedTextWidget) self)->xfwfFormattedText.linegc = XtAllocateGC(self, ((XfwfFormattedTextWidget) self)->core.depth, mask, &values, dynmask, dontcaremask);


    ((XfwfFormattedTextWidget) self)->xfwfFormattedText.last = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.chunks = NULL;
    ((XfwfFormattedTextWidget) self)->xfwfFormattedText.inlined_chunks = NULL;
    ((XfwfFormattedTextWidget) self)->xfwfFormattedText.lines = NULL;
    ((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrlines_alloc = 0;
    ((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftlines = NULL;
    ((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrleftlines_alloc = 0;
    ((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightlines = NULL;
    ((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrrightlines_alloc = 0;
    for (i = 0; i < NR_FAMILIES; i++)
	for (j = 0; j < NR_STYLES; j++)
	    for (k = 0; k < NR_SIZES; k++)
		((XfwfFormattedTextWidget) self)->xfwfFormattedText.fonts[i][j][k].fs = NULL;

    if ((n = sscanf(((XfwfFormattedTextWidget) self)->xfwfFormattedText.fontSizes1, "%d%d%d%d%d%d%d", &((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[0][0],
		    &((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[0][1], &((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[0][2], &((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[0][3],
		    &((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[0][4], &((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[0][5], &((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[0][6])) == 0)
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[0][n++] = 120;	/* Default */
    for (; n < NR_SIZES; n++)
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[0][n] = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[0][n - 1];

    if ((n = sscanf(((XfwfFormattedTextWidget) self)->xfwfFormattedText.fontSizes2, "%d%d%d%d%d%d%d", &((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[1][0],
		    &((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[1][1], &((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[1][2], &((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[1][3],
		    &((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[1][4], &((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[1][5], &((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[1][6])) == 0)
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[1][n++] = 120;	/* Default */
    for (; n < NR_SIZES; n++)
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[1][n] = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[1][n - 1];

    if ((n = sscanf(((XfwfFormattedTextWidget) self)->xfwfFormattedText.fontSizes3, "%d%d%d%d%d%d%d", &((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[2][0],
		    &((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[2][1], &((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[2][2], &((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[2][3],
		    &((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[2][4], &((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[2][5], &((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[2][6])) == 0)
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[2][n++] = 120;	/* Default */
    for (; n < NR_SIZES; n++)
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[2][n] = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[2][n - 1];

    if ((n = sscanf(((XfwfFormattedTextWidget) self)->xfwfFormattedText.fontSizes4, "%d%d%d%d%d%d%d", &((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[3][0],
		    &((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[3][1], &((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[3][2], &((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[3][3],
		    &((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[3][4], &((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[3][5], &((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[3][6])) == 0)
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[3][n++] = 120;	/* Default */
    for (; n < NR_SIZES; n++)
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[3][n] = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[3][n - 1];

    if ((n = sscanf(((XfwfFormattedTextWidget) self)->xfwfFormattedText.fontSizes5, "%d%d%d%d%d%d%d", &((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[4][0],
		    &((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[4][1], &((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[4][2], &((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[4][3],
		    &((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[4][4], &((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[4][5], &((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[4][6])) == 0)
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[4][n++] = 120;	/* Default */
    for (; n < NR_SIZES; n++)
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[4][n] = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fsizes[4][n - 1];

    init_par_builder(self);

    /* $need_reformat = DONT_NEED_REFORMAT; */
    ((XfwfFormattedTextWidget) self)->xfwfFormattedText.workid = 0;
}

static void destroy(Widget self)
{
    FtChunk h;
    int i, j, k;

    if (((XfwfFormattedTextWidget) self)->xfwfFormattedText.workid)
	XtRemoveWorkProc(((XfwfFormattedTextWidget) self)->xfwfFormattedText.workid);
    while (((XfwfFormattedTextWidget) self)->xfwfFormattedText.chunks) {
	h = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.chunks;
	((XfwfFormattedTextWidget) self)->xfwfFormattedText.chunks = h->next;
	if (h->tp == FtWord)
	    XtFree(h->u.word.text);
	XtFree((String) h);
    }
    for (i = 0; i < NR_FAMILIES; i++)
	for (j = 0; j < NR_STYLES; j++)
	    for (k = 0; k < NR_SIZES; k++)
		if (((XfwfFormattedTextWidget) self)->xfwfFormattedText.fonts[i][j][k].fs)
		    XFreeFont(XtDisplay(self), ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fonts[i][j][k].fs);
    XtFree((String) ((XfwfFormattedTextWidget) self)->xfwfFormattedText.lines);
    XtFree((String) ((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftlines);
    XtFree((String) ((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightlines);
}

static void resize(Widget self)
{
    Widget vscroll = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.vscrollbar;

    XtConfigureWidget(vscroll, ((XfwfFormattedTextWidget) self)->core.width - ((XfwfFormattedTextWidget) vscroll)->core.width, 0,
		      ((XfwfFormattedTextWidget) vscroll)->core.width, ((XfwfFormattedTextWidget) self)->core.height, ((XfwfFormattedTextWidget) vscroll)->core.border_width);

    if (((XfwfFormattedTextWidget) self)->xfwfFormattedText.lines || ((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftlines || ((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightlines) {
	/* $need_reformat = NEED_REFORMAT_AND_REDRAW; */
	install_do_layout(self);
    } else
	init_par_builder(self);
}

static void expose(Widget self, XEvent * event, Region region)
{
    int x, y, w, h, i;

    if (!XtIsRealized(self))
	return;

    if (region) {
	XSetRegion(XtDisplay(self), ((XfwfFormattedTextWidget) self)->xfwfFormattedText.gc, region);
	if (((XfwfFormattedTextWidget) self)->xfwfFormattedText.linegc != ((XfwfFormattedTextWidget) self)->xfwfFormattedText.gc)
	    XSetRegion(XtDisplay(self), ((XfwfFormattedTextWidget) self)->xfwfFormattedText.linegc, region);
	if (((XfwfFormattedTextWidget) self)->xfwfFormattedText.spacegc != ((XfwfFormattedTextWidget) self)->xfwfFormattedText.gc)
	    XSetRegion(XtDisplay(self), ((XfwfFormattedTextWidget) self)->xfwfFormattedText.spacegc, region);
    } else {
	XSetClipMask(XtDisplay(self), ((XfwfFormattedTextWidget) self)->xfwfFormattedText.gc, None);
	if (((XfwfFormattedTextWidget) self)->xfwfFormattedText.linegc != ((XfwfFormattedTextWidget) self)->xfwfFormattedText.gc)
	    XSetClipMask(XtDisplay(self), ((XfwfFormattedTextWidget) self)->xfwfFormattedText.linegc, None);
	if (((XfwfFormattedTextWidget) self)->xfwfFormattedText.spacegc != ((XfwfFormattedTextWidget) self)->xfwfFormattedText.gc)
	    XSetClipMask(XtDisplay(self), ((XfwfFormattedTextWidget) self)->xfwfFormattedText.spacegc, None);
    }
    x = event ? event->xexpose.x : 0;
    y = (event ? event->xexpose.y : 0) + ((XfwfFormattedTextWidget) self)->xfwfFormattedText.voffset;
    w = event ? event->xexpose.width : ((XfwfFormattedTextWidget) self)->core.width;
    h = event ? event->xexpose.height : ((XfwfFormattedTextWidget) self)->core.height;

    /* fprintf(stderr, "expose FText (%d,%d,%d,%d)\n", x, y, w, h); */

    i = find_line_below_y(((XfwfFormattedTextWidget) self)->xfwfFormattedText.lines, ((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrlines, y + h) - 1;
    while (i >= 0 && ((XfwfFormattedTextWidget) self)->xfwfFormattedText.lines[i].y + ((XfwfFormattedTextWidget) self)->xfwfFormattedText.lines[i].h > y) {
	draw_chunks(self, ((XfwfFormattedTextWidget) self)->xfwfFormattedText.lines[i].start, ((XfwfFormattedTextWidget) self)->xfwfFormattedText.lines[i].stop);
	i--;
    }
    i = find_line_below_y(((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftlines, ((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrleftlines, y + h) - 1;
    while (i >= 0 && ((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftlines[i].y + ((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftlines[i].h > y) {
	draw_chunks(self, ((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftlines[i].start, ((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftlines[i].stop);
	i--;
    }
    i = find_line_below_y(((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightlines, ((XfwfFormattedTextWidget) self)->xfwfFormattedText.nrrightlines, y + h) - 1;
    while (i >= 0 && ((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightlines[i].y + ((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightlines[i].h > y) {
	draw_chunks(self, ((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightlines[i].start, ((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightlines[i].stop);
	i--;
    }
}

static XtGeometryResult geometry_manager(Widget child, XtWidgetGeometry * request, XtWidgetGeometry * reply)
{
    Widget self = XtParent(child); {
	/* Widget $ = XtParent(child); */
	FtChunk p;
	int newwd, newht;

	/* Find the chunk that contains the widget */
	for (p = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.inlined_chunks; p && p->u.inlined.widget != child;
	     p = p->u.inlined.prev_inlined)
	    assert(p->tp == FtInline);

	if (!p)
	    return XtGeometryYes;	/* Not a chunk */

	if ((request->request_mode & CWWidth)
	    && (p->u.inlined.style & FtALIGNFIXEDWIDTH) == 0) {
	    newwd = request->width;	/* Allow change */
	    p->w = newwd + 2 * p->u.inlined.hmargin;
	    /* if (p->y >= $voffset + $height || p->y + p->h <= $voffset)
	       $need_reformat = NEED_REFORMAT;
	       else
	       $need_reformat = NEED_REFORMAT_AND_REDRAW; */
	    install_do_layout(self);
	} else {
	    newwd = ((XfwfFormattedTextWidget) child)->core.width;	/* Old width */
	}
	if ((request->request_mode & CWHeight)
	    && (p->u.inlined.style & FtALIGNFIXEDHEIGHT) == 0) {
	    newht = request->height;	/* Allow change */
	    p->h = newht - p->d + 2 * p->u.inlined.vmargin;
	    /* if (p->y >= $voffset + $height || p->y + p->h <= $voffset)
	       $need_reformat = NEED_REFORMAT;
	       else
	       $need_reformat = NEED_REFORMAT_AND_REDRAW; */
	    install_do_layout(self);
	} else {
	    newht = ((XfwfFormattedTextWidget) child)->core.height;	/* Old heigth */
	}
	XtResizeWidget(child, newwd, newht, ((XfwfFormattedTextWidget) child)->core.border_width);
	format_inline(self, p);
	return XtGeometryDone;
    }
}

static XtGeometryResult query_geometry(Widget self, XtWidgetGeometry * request, XtWidgetGeometry * reply)
{
    reply->request_mode = CWWidth | CWHeight;
    reply->width = PREF_WD;
    reply->height = PREF_HT;
    if ((request->request_mode & CWWidth) && request->width > PREF_WD)
	reply->width = request->width;
    if ((request->request_mode & CWHeight) && request->height > PREF_HT)
	reply->height = request->height;
    /* fprintf(stderr, "%s: query_geometry %dx%d -> %dx%d\n",
       XtName($), request->width, request->height,
       reply->width, reply->height); */
    if (reply->width == ((XfwfFormattedTextWidget) self)->core.width && reply->height == ((XfwfFormattedTextWidget) self)->core.height)
	return XtGeometryNo;
    if ((request->request_mode & (CWWidth | CWHeight)) != (CWWidth | CWHeight))
	return XtGeometryAlmost;
    if (request->width != reply->width || request->height != reply->height)
	return XtGeometryAlmost;
    return XtGeometryYes;
}

static void add_word(Widget self, const String word, Cardinal len, Pixel fg, Pixel bg, TextStyle style, XtPointer data)
{
    FtChunk p;
    new(p);
    p->tp = FtWord;
#ifndef NDEBUG
    p->x = -9999;
    p->y = -9999;
#endif
    p->u.word.text = strndup(word, len);
    p->u.word.len = len;
    p->u.word.style = style;
    p->u.word.fg = fg;
    p->u.word.bg = bg;
    p->data = data;
    font_extents(self, style, word, len, &p->h, &p->d, &p->w, &p->raise);
    if (style & FtHIDEWIDTH)
	p->w = 0;
    p->u.word.fid = font_id(self, style);
    append(self, p);
}

static void add_hspace(Widget self, int spacefactor, Bool stretchable, Bool breakable, Pixel fg, Pixel bg, TextStyle style, XtPointer data)
{
    FtChunk p;
    new(p);
    p->tp = FtHSpace;
#ifndef NDEBUG
    p->x = -9999;
    p->y = -9999;
    p->w = -9999;		/* Set by render_one_line */
#endif
    font_space(self, style, &p->h, &p->d, &p->u.hspace.normwidth, &p->raise);
    p->u.hspace.normwidth = round(p->u.hspace.normwidth * spacefactor / 1000);
    p->u.hspace.stretchable = stretchable && (spacefactor > 0);
    p->u.hspace.breakable = breakable && (spacefactor > 0);
    if (!p->u.hspace.stretchable)
	p->w = p->u.hspace.normwidth;
    p->u.hspace.fg = fg;
    p->u.hspace.bg = bg;
    p->u.hspace.style = style;
    p->data = data;
    append(self, p);
}

static void add_vspace(Widget self, int pixels)
{
    FtChunk p;

    new(p);
    p->tp = FtVSpace;
#ifndef NDEBUG
    p->x = -9999;
    p->y = -9999;
    p->w = -9999;
    p->d = -9999;
    p->raise = -9999;
    p->data = NULL;
#endif
    p->h = pixels;
    append(self, p);
    format_vspace(self, p);
}

static void add_inline(Widget self, Widget w, TextStyle style, int wd, int ht, int dp, int hmargin, int vmargin, XtPointer data)
{
    FtChunk p;
    int dum1, dum2, dum3;

    new(p);
    p->tp = FtInline;
#ifndef NDEBUG
    p->x = -9999;
    p->y = -9999;
    p->raise = -9999;		/* Set by format_inline */
#endif
    p->w = wd + 2 * hmargin;
    p->h = ht - dp + vmargin;
    p->d = dp + vmargin;
    p->u.inlined.hmargin = hmargin;
    p->u.inlined.vmargin = vmargin;
    p->u.inlined.widget = w;
    p->u.inlined.mapped = FALSE;
    if (XtIsRealized(w)) {
	XtUnmapWidget(w);	/* Mapped by draw_inline() */
	XLowerWindow(XtDisplay(self), XtWindow(w));
    }
    p->u.inlined.style = style;
    p->u.inlined.prev_inlined = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.inlined_chunks;	/* List of inlines */
    ((XfwfFormattedTextWidget) self)->xfwfFormattedText.inlined_chunks = p;
    p->data = data;
    if ((style & (FtALIGNFIXEDWIDTH | FtALIGNFIXEDHEIGHT))) {
	if ((style & FtALIGNFIXEDWIDTH) == 0)
	    wd = ((XfwfFormattedTextWidget) w)->core.width;
	if ((style & FtALIGNFIXEDHEIGHT) == 0)
	    ht = ((XfwfFormattedTextWidget) w)->core.height;
	XtResizeWidget(w, wd, ht, ((XfwfFormattedTextWidget) w)->core.border_width);
    }
    if (style & FtHIDEWIDTH)
	p->w = 0;
    if (data && ((XfwfFormattedTextWidget) self)->xfwfFormattedText.activeCursor && XtIsRealized(w)) {
	XDefineCursor(XtDisplay(w), XtWindow(w), ((XfwfFormattedTextWidget) self)->xfwfFormattedText.activeCursor);
	/* XtAugmentTranslations(w, enterleavetrans); */
    }
    append(self, p);
    format_inline(self, p);
}

static void add_parshape(Widget self, int left, int right, int width, int mode, double leading)
{
    FtChunk p;
    new(p);
    p->tp = FtParShape;
#ifndef NDEBUG
    p->x = -9999;
    p->y = -9999;
    p->w = -9999;
    p->h = -9999;
    p->d = -9999;
    p->raise = -9999;
    p->data = NULL;
#endif
    p->u.parshape.width = width;
    p->u.parshape.left = left;
    p->u.parshape.right = right;
    p->u.parshape.mode = mode;
    p->u.parshape.leading = leading;
    append(self, p);
    format_parshape(self, p);
}

static void add_hrule(Widget self)
{
    FtChunk p;
    new(p);
    p->tp = FtHRule;
#ifndef NDEBUG
    p->x = -9999;
    p->y = -9999;
    p->w = -9999;
    p->raise = -9999;
    p->h = -9999;
    p->data = NULL;
#endif
    p->d = line3d_height;	/* Height of bitmap */
    append(self, p);
    format_hrule(self, p);
}

static void add_eod(Widget self)
{
    ((XfwfFormattedTextWidgetClass) self->core.widget_class)->xfwfFormattedText_class.add_vspace(self, 0);
    /*dump(self);*/
}

static void pass_click(Widget self, Widget child, int x, int y)
{
    XfwfFTextCallbackStruct info;
    FtChunk p;

    for (p = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.inlined_chunks; p; p = p->u.inlined.prev_inlined)
	if (p->u.inlined.widget == child) {
	    if (p->data) {
		info.reason = XfwfActivate;
		info.event = NULL;
		info.x = x;
		info.y = y;
		info.data = p->data;
		XtCallCallbackList(self, ((XfwfFormattedTextWidget) self)->xfwfFormattedText.activate, &info);
	    }
	    break;
	}
}

static void reformat_scrollbar(Widget self)
{
    int value, slider, maximum, pageinc;

    XtVaGetValues(((XfwfFormattedTextWidget) self)->xfwfFormattedText.vscrollbar, XmNvalue, &value, NULL);
    maximum = max(((XfwfFormattedTextWidget) self)->xfwfFormattedText.ypos, max(((XfwfFormattedTextWidget) self)->xfwfFormattedText.leftypos, ((XfwfFormattedTextWidget) self)->xfwfFormattedText.rightypos));
    set_max(maximum, 1);
    slider = min(((XfwfFormattedTextWidget) self)->core.height, maximum);
    set_min(value, maximum - slider);
    ((XfwfFormattedTextWidget) self)->xfwfFormattedText.voffset = value;
    pageinc = max(1, ((XfwfFormattedTextWidget) self)->core.height - 10);
    XtVaSetValues(((XfwfFormattedTextWidget) self)->xfwfFormattedText.vscrollbar, XmNmaximum, maximum, XmNsliderSize, slider,
		  XmNvalue, value, XmNincrement, 10,
		  XmNpageIncrement, pageinc, NULL);
}

static int get_em_of_textstyle(Widget self, TextStyle style)
{
    int fam, sty, siz, raise;
    interpret_style(self, style, &fam, &sty, &siz, &raise);
    return ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fonts[fam][sty][siz].em;
}

static int get_space_of_textstyle(Widget self, TextStyle style)
{
    int fam, sty, siz, raise;
    interpret_style(self, style, &fam, &sty, &siz, &raise);
    return ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fonts[fam][sty][siz].space;
}

static void get_lineheight_of_textstyle(Widget self, TextStyle style, int *h, int *d)
{
    int fam, sty, siz, raise;
    interpret_style(self, style, &fam, &sty, &siz, &raise);
    *h = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fonts[fam][sty][siz].fs->ascent;
    *d = ((XfwfFormattedTextWidget) self)->xfwfFormattedText.fonts[fam][sty][siz].fs->descent;
}

static void get_word_extent(Widget self, const String word, Cardinal len, TextStyle textstyle, int *h, int *d, int *w, int *raise)
{
    font_extents(self, textstyle, word, len, h, d, w, raise);
}

void XfwfAddWord(Widget self, const String word, Cardinal len, Pixel fg, Pixel bg, TextStyle style, XtPointer data)
{
    assert(XtIsSubclass(self, xfwfFormattedTextWidgetClass));
    ((XfwfFormattedTextWidgetClass) self->core.widget_class)->xfwfFormattedText_class.add_word(self, word, len, fg, bg, style, data);
}

void XfwfAddHSpace(Widget self, int spacefactor, Bool stretchable, Bool breakable, Pixel fg, Pixel bg, TextStyle style, XtPointer data)
{
    assert(XtIsSubclass(self, xfwfFormattedTextWidgetClass));
    ((XfwfFormattedTextWidgetClass) self->core.widget_class)->xfwfFormattedText_class.add_hspace(self, spacefactor, stretchable, breakable, fg, bg, style, data);
}

void XfwfAddVSpace(Widget self, int pixels)
{
    assert(XtIsSubclass(self, xfwfFormattedTextWidgetClass));
    ((XfwfFormattedTextWidgetClass) self->core.widget_class)->xfwfFormattedText_class.add_vspace(self, pixels);
}

void XfwfAddInline(Widget self, Widget w, TextStyle style, int wd, int ht, int dp, int hmargin, int vmargin, XtPointer data)
{
    assert(XtIsSubclass(self, xfwfFormattedTextWidgetClass));
    ((XfwfFormattedTextWidgetClass) self->core.widget_class)->xfwfFormattedText_class.add_inline(self, w, style, wd, ht, dp, hmargin, vmargin, data);
}

void XfwfAddParShape(Widget self, int left, int right, int width, int mode, double leading)
{
    assert(XtIsSubclass(self, xfwfFormattedTextWidgetClass));
    ((XfwfFormattedTextWidgetClass) self->core.widget_class)->xfwfFormattedText_class.add_parshape(self, left, right, width, mode, leading);
}

void XfwfAddHRule(Widget self)
{
    assert(XtIsSubclass(self, xfwfFormattedTextWidgetClass));
    ((XfwfFormattedTextWidgetClass) self->core.widget_class)->xfwfFormattedText_class.add_hrule(self);
}

void XfwfAddEOD(Widget self)
{
    assert(XtIsSubclass(self, xfwfFormattedTextWidgetClass));
    ((XfwfFormattedTextWidgetClass) self->core.widget_class)->xfwfFormattedText_class.add_eod(self);
}

void XfwfPassClick(Widget self, Widget child, int x, int y)
{
    assert(XtIsSubclass(self, xfwfFormattedTextWidgetClass));
    ((XfwfFormattedTextWidgetClass) self->core.widget_class)->xfwfFormattedText_class.pass_click(self, child, x, y);
}
