/*
 *    Copyright  1996, 1997, 1998, 1999, 2000 by Remco Treffkorn
 *    Distributed under the terms of the Q Public License. Please see
 *    the included file LICENSE.
 */
#include "config.h"

#include <stdio.h>

#include <Xm/DialogS.h>
#include <Xm/ScrolledW.h>
#include <Xm/DrawingA.h>
#include <Xm/PushB.h>
#include <Xm/TextF.h>
#include <Xm/Label.h>
#include <Xm/Frame.h>
#include <Xm/BulletinB.h>
#include "xmap.h"
#include "gps.h"
#include "values.h"
#include "debug.h"

static Widget monitordialog = NULL;
static Widget bul;
static Widget start;
static Widget cancel;
static Widget aunit, sunit;
static Widget status;
static Widget latitude, longitude, altitude, headingw, velocity, date,
 wtime;


extern XtAppContext app;
static XtIntervalId id;


enum {
    SUNKNOWN, SUNCONNECTED, SCONNECTED
};
static int last_status = SUNKNOWN;
static int last_unit = UNKNOWN;
static int last_gpsstatus = -1;
static int last_gpsmode = -1;

static double last_latitude, last_longitude, last_altitude;
static double last_heading, last_velocity;

static char last_date[32];
static char last_time[32];
static time_t last_gpstime;
static int oldformat;
static Boolean process = False;


struct c_values cv;

void update_monitor_units(void)
{
    XmString s;

    switch (unit) {
    case ENGLISH:
	cv.altitude = "feet";
	cv.velocity = "mph";
	cv.distunit = "mi";
	break;
    case METRIC:
	cv.altitude = "meters";
	cv.velocity = "kmh";
	cv.distunit = "Km";
	break;
    case NAUTICAL:
	cv.altitude = "feet";
	cv.velocity = "knots";
	cv.distunit = "nm";
	break;
    }
    cv.update |= U_UNITS;
    last_unit = unit;
}

static void update_field(Widget w, double new, char *format)
{
    char bufp[128];

    snprintf(bufp, 127, format, new);
    XtVaSetValues(w, XmNvalue, bufp, NULL);
}

static void update_heading()
{
    if (last_longitude != gpsdata.longitude
	|| last_latitude != gpsdata.latitude) {
	cv.heading_val = heading(last_latitude, last_longitude,
				 gpsdata.latitude, gpsdata.longitude);
	cv.update |= U_HEADING;
    }
}

static void update_lat(double new, double *old)
{
    if (new != *old || oldformat != options.format) {
	cv.update |= U_LATITUDE;
	cv.latitude = new;
	*old = new;
    }
}

static void update_lon(double new, double *old)
{
    if (new != *old || oldformat != options.format) {
	cv.update |= U_LONGITUDE;
	cv.longitude = new;
	*old = new;
    }
}

void do_conversions()
{
    XmString s;
    char *m;
    double val;

    if (last_unit != unit)
	update_monitor_units();

    update_heading();
    update_lat(gpsdata.latitude, &last_latitude);
    update_lon(gpsdata.longitude, &last_longitude);

    if (unit == ENGLISH || unit == NAUTICAL)
	cv.altitude_val = gpsdata.altitude / 0.3048;
    else
	cv.altitude_val = gpsdata.altitude;

    if (unit == METRIC)
	cv.velocity_val = 1.852 * gpsdata.velocity;
    else if (unit == ENGLISH)
	cv.velocity_val = 1.852 / 1.609347 * gpsdata.velocity;
    else
	cv.velocity_val = gpsdata.velocity;


    if (cv.altitude_val != last_altitude) {
	cv.update |= U_ALTITUDE;
	last_altitude = cv.altitude_val;
    }
    if (cv.velocity_val != last_velocity) {
	cv.update |= U_VELOCITY;
	last_velocity = cv.velocity_val;
    }
    if (strcmp(gpsdata.date, last_date)) {
	strcpy(last_date, gpsdata.date);
	cv.update |= U_DATE;
    }
    if (strcmp(gpsdata.time, last_time)) {
	strcpy(last_time, gpsdata.time);
	cv.update |= U_TIME;
    }
    if (gpsdata.gpstime != last_gpstime) {
	last_gpstime = gpsdata.gpstime;
	cv.update |= (U_TIME | U_DATE);
    }
    if (gpsdata.status > -1) {	// Connected

	if (last_gpsstatus != gpsdata.status || last_gpsmode != gpsdata.mode) {
	    switch (gpsdata.status) {
	    case 0:
		m = "Initializing...";
		break;
	    case 1:
	    case 2:
		switch (gpsdata.mode) {
		case 0:	// this really only means this field is not valid!

		    m = "Locked on";
		    break;
		case 1:
		    m = "No fix, data invalid";
		    break;
		case 2:
		    m = "2D fix";
		    break;
		case 3:
		    m = "3D fix";
		    break;
		}
		break;
	    default:
		m = "messed up";
	    }
	    cv.status = m;
	    cv.update |= U_STATUS;
	    last_gpsstatus = gpsdata.status;
	    last_gpsmode = gpsdata.mode;
	}
	if (last_status != SCONNECTED) {
	    cv.ss_button_label = "Stop";
	    cv.update |= U_LABEL;
	    last_status = SCONNECTED;
	}
    } else {			// Not Connected

	if (gpsdata.status == CONNECT_FAILED) {
	    cv.status = "Unable to connected to GPS server";
	    cv.update |= U_STATUS;

	    cv.ss_button_label = "Start";
	    cv.update |= U_LABEL;

	    gpsdata.status = UNCONNECTED;
	    last_gpsstatus = gpsdata.status;
	    last_status = SUNCONNECTED;
	}
	if (last_status != SUNCONNECTED) {
	    cv.status = "Not connected to GPS server";
	    cv.ss_button_label = "Start";
	    cv.update |= U_LABEL;

	    last_gpsstatus = gpsdata.status;
	    last_status = SUNCONNECTED;
	}
    }
}

int gps_data_valid()
{
    if (gpsdata.status > 0)
	if (gpsdata.mode == 0 || gpsdata.mode == 2 || gpsdata.mode == 3)
	    return 1;
    return 0;
}

void finish_updates()
{
    cv.update = 0;
}


void do_monitor()
{
    XmString s;
    char *m;
    double val;
    char bufp[256];

    if (!process)
	return;

    if (cv.update & U_LABEL) {
	s = XmStringCreateSimple(cv.ss_button_label);
	XtVaSetValues(start,
		      XmNlabelString, s,
		      NULL);
	XmStringFree(s);
    }
    if (cv.update & U_UNITS) {
	s = XmStringCreateSimple(cv.altitude);
	XtVaSetValues(aunit,
		      XmNlabelString, s,
		      NULL);
	XmStringFree(s);
	s = XmStringCreateSimple(cv.velocity);
	XtVaSetValues(sunit,
		      XmNlabelString, s,
		      NULL);
	XmStringFree(s);
    }
    if (cv.update & U_STATUS) {
	XtVaSetValues(status,
		      XmNvalue, cv.status,
		      NULL);
    }
    if (cv.update & U_HEADING) {
	snprintf(bufp, 255, "% 3.0lf", cv.heading_val);
	XtVaSetValues(headingw, XmNvalue, bufp, NULL);
    }
    if (cv.update & U_LATITUDE) {
	toUserFormatLat(options.format, gpsdata.latitude, bufp, 255);
	XtVaSetValues(latitude, XmNvalue, bufp, NULL);
    }
    if (cv.update & U_LONGITUDE) {
	toUserFormatLon(options.format, gpsdata.longitude, bufp, 255);
	XtVaSetValues(longitude, XmNvalue, bufp, NULL);
    }
    if (cv.update & U_ALTITUDE)
	update_field(altitude, cv.altitude_val, "% 5.0lf");

    if (cv.update & U_VELOCITY)
	update_field(velocity, cv.velocity_val, "% 5.1lf");


    if (cv.update & (U_DATE | U_TIME) && gpsdata.gpstime != -1) {
	char tbuf[20];

	if (options.useutc) {
	    strftime(tbuf, 20, "%x", gmtime(&gpsdata.gpstime));
	    XtVaSetValues(date, XmNvalue, tbuf, NULL);
	    strftime(tbuf, 20, "%X", gmtime(&gpsdata.gpstime));
	    XtVaSetValues(wtime, XmNvalue, tbuf, NULL);
	} else {
	    strftime(tbuf, 20, "%x", localtime(&gpsdata.gpstime));
	    XtVaSetValues(date, XmNvalue, tbuf, NULL);
	    strftime(tbuf, 20, "%X", localtime(&gpsdata.gpstime));
	    XtVaSetValues(wtime, XmNvalue, tbuf, NULL);
	}
    } else {
	if (cv.update & U_DATE)
	    XtVaSetValues(date, XmNvalue, gpsdata.date, NULL);
	if (cv.update & U_TIME)
	    XtVaSetValues(wtime, XmNvalue, gpsdata.time, NULL);
    }
}

static void ss_cb(Widget w, XtPointer client_data, XmAnyCallbackStruct * cbs)
{
    track_toggle();
} static void continue_cb(Widget w, XtPointer client_data, XmAnyCallbackStruct * cbs)
{
    if (id)
	XtRemoveTimeOut(id);
    id = 0;
    XtPopdown(monitordialog);
}
void do_monitor_dialog(Widget w, XtPointer client_data, XtPointer call_data)
{
    Arg args[100];
    int n;
    Pixel foreground;

    last_unit = UNKNOWN;

    if (!monitordialog) {

	XmString s;

	s = XmStringCreateSimple("");
	foreground = get_pixel(w, "Black");
	monitordialog = XtVaCreateWidget("monitordialog",
				xmDialogShellWidgetClass, GetTopShell(w),
				      XmNtitle, "Monitor GPS Connection",	/* give arbitrary title in wm */
					 XmNdeleteResponse, XmUNMAP,	/* system menu "Close" action */
					 XmNresizePolicy, XmRESIZE_NONE,
					 XmNnoResize, True,
					 XmNwidth, 300,
					 XmNminWidth, 300,
					 XmNmaxWidth, 300,
					 XmNheight, 200,
					 XmNminHeight, 200,
					 XmNmaxHeight, 200,
					 XmNmarginWidth, 5,
					 XmNmarginHeight, 5,
					 NULL);

	bul = XtVaCreateManagedWidget("bul",
			       xmBulletinBoardWidgetClass, monitordialog,
				      XmNforeground, foreground,
				      XmNautoUnmanage, False,
				      XmNdialogStyle, XmDIALOG_WORK_AREA,
				      XmNwidth, 300,
				      XmNheight, 200,
				      NULL);
	XtAddCallback(bul, XmNunmapCallback, (void *) continue_cb, NULL);
//      XtManageChild(bul);

	XtVaCreateManagedWidget("Status:", xmLabelWidgetClass, bul,
				XmNx, 10,
				XmNy, 10,
				NULL);

	status = XtVaCreateManagedWidget("status", xmTextFieldWidgetClass, bul,
					 XmNx, 60,
					 XmNy, 10,
					 XmNwidth, 225,
					 XmNcursorPositionVisible, False,
					 XmNeditable, False,
					 XmNmarginHeight, 2,
					 XmNhighlightThickness, 0,
					 XmNshadowThickness, 0,
					 NULL);

#if 0
	XtVaCreateManagedWidget("sfield", xmLabelWidgetClass, bul,
				XmNlabelString, s,
				XmNx, 60,
				XmNy, 10,
				NULL);
#endif

	XtVaCreateManagedWidget("fr1", xmFrameWidgetClass, bul,
				XmNx, 0,
				XmNy, 40,
				XmNwidth, 160,
				XmNheight, 150,
				NULL);

	XtVaCreateManagedWidget("fr2", xmFrameWidgetClass, bul,
				XmNx, 180,
				XmNy, 40,
				XmNwidth, 110,
				XmNheight, 70,
				NULL);


	cancel = XtVaCreateManagedWidget("Continue",
					 xmPushButtonWidgetClass, bul,
					 XmNforeground, foreground,
					 XmNhighlightThickness, 0,
					 XmNwidth, 60,
					 XmNx, 210,
					 XmNy, 145,
					 NULL);
	XtAddCallback(cancel, XmNactivateCallback, (void *) continue_cb, NULL);

	start = XtVaCreateManagedWidget("X",
					xmPushButtonWidgetClass, bul,
					XmNforeground, foreground,
					XmNhighlightThickness, 0,
					XmNwidth, 60,
					XmNx, 210,
					XmNy, 170,
					NULL);
	XtAddCallback(start, XmNactivateCallback, (void *) ss_cb, NULL);


// do the labels

#define ROFF 55
#define SPACING 27

#define ROW(x) (ROFF + (x * SPACING))

#define COL1 13
#define COL2 80
#define COL3 128
#define COL4 183
#define COL5 218

	XtVaCreateManagedWidget("Location", xmLabelWidgetClass, bul,
				XmNx, COL1 + 1,
				XmNy, 32,
				NULL);

	XtVaCreateManagedWidget("Date/Time", xmLabelWidgetClass, bul,
				XmNx, COL4 + 1,
				XmNy, 32,
				NULL);

	XtVaCreateManagedWidget("Latitude:", xmLabelWidgetClass, bul,
				XmNx, COL1,
				XmNy, ROW(0),
				NULL);

	XtVaCreateManagedWidget("Date:", xmLabelWidgetClass, bul,
				XmNx, COL4,
				XmNy, ROW(0),
				NULL);

	XtVaCreateManagedWidget("Longitude:", xmLabelWidgetClass, bul,
				XmNx, COL1,
				XmNy, ROW(1),
				NULL);

	XtVaCreateManagedWidget("Time:", xmLabelWidgetClass, bul,
				XmNx, COL4,
				XmNy, ROW(1),
				NULL);

	XtVaCreateManagedWidget("Elevation:", xmLabelWidgetClass, bul,
				XmNx, COL1,
				XmNy, ROW(2),
				NULL);

	aunit = XtVaCreateManagedWidget("a", xmLabelWidgetClass, bul,
					XmNlabelString, s,
					XmNx, COL3,
					XmNy, ROW(2),
					NULL);

	XtVaCreateManagedWidget("Heading:", xmLabelWidgetClass, bul,
				XmNx, COL1,
				XmNy, ROW(3),
				NULL);

	XtVaCreateManagedWidget("Speed:", xmLabelWidgetClass, bul,
				XmNx, COL1,
				XmNy, ROW(4),
				NULL);

	sunit = XtVaCreateManagedWidget("s", xmLabelWidgetClass, bul,
					XmNlabelString, s,
					XmNx, COL3,
					XmNy, ROW(4),
					NULL);


	XtVaCreateManagedWidget("t", xmLabelWidgetClass, bul,
				XmNx, 205,
				XmNy, 120,
				XmNlabelString, s,
				NULL);

	XmStringFree(s);

// do the text fields

	latitude = XtVaCreateManagedWidget("lat", xmTextFieldWidgetClass, bul,
					   XmNx, COL2,
					   XmNy, ROW(0),
					   XmNwidth, 87,
					 XmNcursorPositionVisible, False,
					   XmNeditable, False,
					   XmNmarginHeight, 2,
					   XmNhighlightThickness, 0,
					   XmNshadowThickness, 0,
					   NULL);

	longitude = XtVaCreateManagedWidget("lon", xmTextFieldWidgetClass, bul,
					    XmNx, COL2,
					    XmNy, ROW(1),
					    XmNwidth, 87,
					 XmNcursorPositionVisible, False,
					    XmNeditable, False,
					    XmNmarginHeight, 2,
					    XmNhighlightThickness, 0,
					    XmNshadowThickness, 0,
					    NULL);

	altitude = XtVaCreateManagedWidget("elev", xmTextFieldWidgetClass, bul,
					   XmNx, COL2,
					   XmNy, ROW(2),
					   XmNwidth, 50,
					 XmNcursorPositionVisible, False,
					   XmNeditable, False,
					   XmNmarginHeight, 2,
					   XmNhighlightThickness, 0,
					   XmNshadowThickness, 0,
					   NULL);

	headingw = XtVaCreateManagedWidget("head", xmTextFieldWidgetClass, bul,
					   XmNx, COL2,
					   XmNy, ROW(3),
					   XmNwidth, 50,
					 XmNcursorPositionVisible, False,
					   XmNeditable, False,
					   XmNmarginHeight, 2,
					   XmNhighlightThickness, 0,
					   XmNshadowThickness, 0,
					   NULL);

	velocity = XtVaCreateManagedWidget("vel", xmTextFieldWidgetClass, bul,
					   XmNx, COL2,
					   XmNy, ROW(4),
					   XmNwidth, 50,
					 XmNcursorPositionVisible, False,
					   XmNeditable, False,
					   XmNmarginHeight, 2,
					   XmNhighlightThickness, 0,
					   XmNshadowThickness, 0,
					   NULL);

	date = XtVaCreateManagedWidget("dat", xmTextFieldWidgetClass, bul,
				       XmNx, COL5 - 5,
				       XmNy, ROW(0),
				       XmNwidth, 72,
				       XmNcursorPositionVisible, False,
				       XmNeditable, False,
				       XmNmarginHeight, 2,
				       XmNhighlightThickness, 0,
				       XmNshadowThickness, 0,
				       NULL);

	wtime = XtVaCreateManagedWidget("tim", xmTextFieldWidgetClass, bul,
					XmNx, COL5,
					XmNy, ROW(1),
					XmNwidth, 58,
					XmNcursorPositionVisible, False,
					XmNeditable, False,
					XmNmarginHeight, 2,
					XmNhighlightThickness, 0,
					XmNshadowThickness, 0,
					NULL);
    }
    XtManageChild(bul);
    XtManageChild(monitordialog);
    XtPopup(monitordialog, XtGrabNone);
    last_status = SUNKNOWN;

    process = True;
}
